import pysam

def getRealReadPos(read, offset):

        if read.is_reverse:
            return getBackwardReadPos(read, offset)
        else:
            return getForwardReadPos(read, offset)+1


def getForwardReadPos(read, offset):
    sum = 0
    skip = 0
    indices = range(len(read.cigar))
    pos = read.pos
    for i in indices:
        t = read.cigar[i]

        nextStep = t[1]
        if t[0] == 3 :
            skip += nextStep
        else:
            if sum + nextStep > -offset:
                break
            sum += nextStep

    pos += (-offset + skip)
    return pos

def getBackwardReadPos(read, offset):
    sum = 0
    skip = 0
    indices = range(len(read.cigar)-1,-1,-1)

    pos = read.pos + read.alen

    for i in indices:

        t = read.cigar[i]
        nextStep = t[1]

        if t[0] == 3 :
            skip += nextStep
        else:
            if sum + nextStep > -offset:
                break
            sum += nextStep

    #if(offset!=0):
    #    print(read.cigar)
    pos -= (-offset + skip)

    #print "offset and skip: ", offset , skip
    return pos

def parseRead(read):



        if len(read.cigar) == 1:
            return read.alen, [(read.pos, read.pos+read.alen)], []

        length = 0
        gaps = []
        fills =[]
        currentPos = read.pos

        for t in read.cigar:
            prevPos = currentPos
            currentPos += t[1]
            if t[0] != 3:
                length += t[1]
                fills.append((prevPos, currentPos))
            else:
                gaps.append((prevPos, currentPos))
            currentPos += 1


        return length, fills, gaps

def isMultiMapper(read):

    tags = dict(read.tags)
    if not 'NH' in tags:
        return False
    return tags['NH'] > 1
##############################################################################################

def getPosOnExon(pos, transcript):


    if not (transcript['RNA_min'] <= pos <= transcript['RNA_max']):
        return -2

    for i in range(0, len(transcript['exon_positions'])):
        exon = transcript['exon_positions'][i]
        size = transcript['exon_sizes'][i]
        interval = (exon, exon + size)

        if interval[0] <= pos <= interval[1]:
            return i
    #print(pos, transcript['exon_positions'], transcript['exon_sizes'])
    return -1

def getOnExon(read, transcript, offset):

    realPos = getRealReadPos(read, offset)
    return getPosOnExon(realPos, transcript)

def checInterval(interval, pos):
    if interval[0] <= pos <= interval[1]:
        return 0
    elif pos < interval[0]:
        return -1
    elif interval[1] < pos:
        return 1

def dnatotranscript(dnapos, transcript):
    transcriptpos = 0
    exonIndex = -1
    for i in range(0, len(transcript['exon_positions'])):
        exon = transcript['exon_positions'][i]
        size = transcript['exon_sizes'][i]
        interval = (exon, exon + size)

        if interval[0] <= dnapos <= interval[1]:
            exonIndex = i
            transcriptpos += dnapos - exon
            break
        transcriptpos += size

    if exonIndex < 0:
        return -1
    return transcriptpos

def transcripttodna(transcriptpos, transcript):

    for i in range(0, len(transcript['exon_positions'])):
        exon = transcript['exon_positions'][i]
        size = transcript['exon_sizes'][i]


        next = transcriptpos - size
        if next <= 0:
            return exon + transcriptpos
            break
        transcriptpos -= size

    return -1

def moveOnTranscript(pos, transcript, steps) :

    transcriptpos = dnatotranscript(pos, transcript)
    transcriptpos += steps
    dnapos = transcripttodna(transcriptpos, transcript)

    #if abs(dnapos - pos) != 1:
    #    print pos, dnapos, transcript['transcript_id']
    return dnapos

def getDiffToCDS(read,transcript, offset,compareToMin  = True):
    if compareToMin:
        return getDiffToMinCDS(read, transcript, offset)
    else:
        return getDiffToMaxCDS(read, transcript, offset)
def getDiffToMinCDS(read, transcript, offset):
    realPos = getRealReadPos(read, offset)

    currentExon =getOnExon(read, transcript, offset)
    if currentExon < 0:
        return 4000

    differences = transcript['exon_distances'][currentExon]
    exonPos = transcript['exon_positions'][currentExon]


    diff = 0
    pos =transcript['CDS_min']
    if not transcript['forward']:
        pos =transcript['stop_codon'][1]

    if currentExon == transcript['start_exon']:
        diff = realPos - pos
    else:
        diff = differences[0] + realPos-exonPos

    if not transcript['forward']:
        diff = -diff


    return diff

def getDiffToMaxCDS(read, transcript, offset):
    realPos = getRealReadPos(read, offset)

    currentExon =getOnExon(read, transcript, offset)

    if currentExon < 0:
        return 4000

    differences = transcript['exon_distances'][currentExon]
    exonPos = transcript['exon_positions'][currentExon]
    exonSize = transcript['exon_sizes'][currentExon]

    diff = 0
    pos =transcript['CDS_max']

    if transcript['forward']:
        pos =transcript['stop_codon'][0]




    if currentExon == transcript['start_exon']:
        diff = pos - realPos

    else:
        diff = differences[1] + (exonSize - (realPos-exonPos))

    if  transcript['forward']:
             diff = -diff

    return diff



def getModulatedReadPos(read, gene, offset):

        modPos =  getDiffToCDS(read, gene, offset)

        return modPos % 3



from OpenGL.GL import *
from OpenGL.GLU import *
import numpy as np
from OpenGL.GLUT import *
from PyQt5.Qt import *
from PyQt5 import QtCore
from shoelaces.resource import resourceItemDelegate
from ctypes import *
import datetime
import math

import shoelaces.settings as settings


class exportDialog(QDialog):
    def __init__(self, offsets, parent=None):
        super(exportDialog, self).__init__(parent)

        self.centerWidget = self
        self.mainLayout = QGridLayout(self.centerWidget)
        self.setLayout(self.mainLayout)

        self.outputType = QComboBox(self.centerWidget)
        self.outputType.addItems(['Single File', 'Per footprint length'])
        self.offsetBox = QComboBox(self.centerWidget)
        self.offsetBox.addItems(['Orginal', 'Use Offsets','Per transcript', 'Per transcript w/offsets'])
        self.offsetBox.setCurrentIndex(1)
        self.mmBox = QComboBox(self.centerWidget)
        self.mmBox.addItems(['Ignore', 'Merge', 'Split'])

        self.fileNameInput = QLineEdit(self.centerWidget)
        self.fileNameButton =  QPushButton('...', self.centerWidget)

        self.okButton = QPushButton("OK",self)
        self.cancelButton = QPushButton("Cancel",self)

        self.offsetBoxes = {}

        for idx, key in enumerate(offsets):
            label = QLabel(str(key) + ': ' + str(offsets[key]), self.centerWidget)
            checkBox = QCheckBox(self.centerWidget)
            self.mainLayout.addWidget(label, 2,idx)
            self.mainLayout.addWidget(checkBox, 3,idx)
            self.offsetBoxes[key] = checkBox

        self.mainLayout.addWidget(QLabel("Output:"), 0, 0)
        self.mainLayout.addWidget(self.outputType, 1, 0)
        self.mainLayout.addWidget(QLabel("Position:"), 0, 1)
        self.mainLayout.addWidget(self.offsetBox, 1, 1)
        self.mainLayout.addWidget(QLabel("Multimappers:"), 0, 2)
        self.mainLayout.addWidget(self.mmBox, 1, 2)

        self.mainLayout.addWidget(QLabel("Filename:"), 4, 0 )
        self.mainLayout.addWidget(self.fileNameInput, 4, 1, 1, 1)
        self.mainLayout.addWidget(self.fileNameButton, 4, 2)
        self.mainLayout.addWidget(self.okButton, 5, 1)
        self.mainLayout.addWidget(self.cancelButton, 5, 2)
        self.offsets = offsets
        self.adjustSize()


        self.okButton.pressed.connect(self.accept)
        self.cancelButton.pressed.connect(self.reject)
        self.fileNameButton.pressed.connect(self.browseFileName)

    def getConfig(self):
        lengths = []
        for length, box in self.offsetBoxes.items():
            if box.checkState():
                lengths.append(length)

        useOffset = self.offsetBox.currentText() == 'Use Offsets' or self.offsetBox.currentText() == 'Per transcript w/offsets'
        relative = self.offsetBox.currentText() == 'Per transcript' or self.offsetBox.currentText() == 'Per transcript w/offsets'
        split = self.outputType.currentText() == 'Per footprint length'
        filename = str(self.fileNameInput.text())
        multimappers = self.mmBox.currentText()
        return useOffset, lengths, filename, split, multimappers, relative

    def browseFileName(self):
        filename, _ = QFileDialog.getSaveFileName(self, "Save Wig File", '',
                    "Wig files (*.wig)")
        self.fileNameInput.setText(filename)

class batchDialog(QDialog):
    def __init__(self, resources,  offsets, parent=None):
        super(batchDialog, self).__init__(parent)

        self.centerWidget = self
        self.mainLayout = QGridLayout(self.centerWidget)
        self.setLayout(self.mainLayout)

        self.outputType = QComboBox(self.centerWidget)
        self.outputType.addItems(['Single File', 'Per footprint length'])
        self.offsetBox = QComboBox(self.centerWidget)
        self.offsetBox.addItems(['Orginal', 'Use Offsets','Per transcript', 'Per transcript w/offsets'])
        self.offsetBox.setCurrentIndex(1)
        self.mmBox = QComboBox(self.centerWidget)
        self.mmBox.addItems(['Output Wig'])


        self.fileNameInput = QComboBox(self.centerWidget)
        transcriptFiles = []
        self.fileNames = []
        for res in resources:
            if res.GetType() == 'TranscriptResource':
                transcriptFiles.append(res.GetName())
            elif res.GetType() == 'SequenceResource':
                self.fileNames.append(res.GetName())
        self.fileNameInput.addItems(transcriptFiles)

        self.fileNameButton = QPushButton('...', self.centerWidget)

        self.fileNameView = QTreeWidget(self.centerWidget)
        self.fileNameView.setColumnCount(2)
        self.fileNameView.setHeaderLabels(['Include', 'Name'])

        for name in self.fileNames:
            item = QTreeWidgetItem()
            item.setData(1, Qt.DisplayRole, name)

            item.setCheckState(0,Qt.Checked);
            self.fileNameView.addTopLevelItem(item)




        self.addFileButton = QPushButton('Add', self.centerWidget)

        self.directoryOutput = QLineEdit(self.centerWidget)
        self.directoryOutputButton = QPushButton('...', self.centerWidget)

        self.offsetView = QTreeWidget(self.centerWidget)
        self.offsetView.setColumnCount(3)
        self.offsetView.setHeaderLabels(['Include', 'Length', 'Offset'])


        self.okButton = QPushButton("OK",self)
        self.cancelButton = QPushButton("Cancel",self)



        for key, value in offsets.items():
            item = QTreeWidgetItem()
            item.setData(1, Qt.DisplayRole, key)
            item.setData(2, Qt.DisplayRole, value)
            item.setCheckState(0,Qt.Checked);

            self.offsetView.addTopLevelItem(item)


        self.mainLayout.addWidget(QLabel("Output:"), 1, 0)
        self.mainLayout.addWidget(self.outputType, 1, 1)
        self.mainLayout.addWidget(QLabel("Position:"), 1, 2)
        self.mainLayout.addWidget(self.offsetBox, 1, 3)
        #self.mainLayout.addWidget(QLabel("Multimappers:"), 3, 0)
        self.mainLayout.addWidget(self.mmBox, 0,0,1,4)


        self.mainLayout.addWidget(QLabel("Data Files"), 3, 2)
        self.mainLayout.addWidget(self.addFileButton, 3, 3)

        self.mainLayout.addWidget(self.fileNameView, 4, 2, 1, 2)

        self.mainLayout.addWidget(QLabel("Transcript File:"), 2, 0)
        self.mainLayout.addWidget(self.fileNameInput, 2,1, 1, 2)
        self.mainLayout.addWidget(self.fileNameButton, 2, 3)
        self.mainLayout.addWidget(QLabel("Offsets"), 3, 0, 1, 2)
        self.mainLayout.addWidget(self.offsetView, 4, 0, 1, 2)

        self.mainLayout.addWidget(QLabel("Output Directory:"), 5, 0)
        self.mainLayout.addWidget(self.directoryOutput, 5,1, 1, 2)
        self.mainLayout.addWidget(self.directoryOutputButton, 5, 3)


        self.mainLayout.addWidget(self.okButton, 6, 2)
        self.mainLayout.addWidget(self.cancelButton, 6, 3)
        self.offsets = offsets
        self.adjustSize()


        self.okButton.pressed.connect(self.accept)
        self.cancelButton.pressed.connect(self.reject)
        self.fileNameButton.pressed.connect(self.browseTranscriptFileName)
        self.addFileButton.pressed.connect(self.browseFileName)
        self.directoryOutputButton.pressed.connect(self.browseDirectory)

    def getConfig(self):

        lengths = []
        itemIterator =  QTreeWidgetItemIterator(self.offsetView, QTreeWidgetItemIterator.Checked)
        while itemIterator.value():
            item = itemIterator.value()
            key = eval(str(item.data(1,Qt.DisplayRole)))
            lengths.append(key)
            itemIterator += 1


        itemIterator =  QTreeWidgetItemIterator(self.fileNameView, QTreeWidgetItemIterator.Checked)
        dataFiles = []
        while itemIterator.value():


            item = itemIterator.value()

            name = str(item.data(1,Qt.DisplayRole))

            dataFiles.append(name)

            itemIterator += 1

        useOffset = self.offsetBox.currentText() == 'Use Offsets' or self.offsetBox.currentText() == 'Per transcript w/offsets'
        relative = self.offsetBox.currentText() == 'Per transcript' or self.offsetBox.currentText() == 'Per transcript w/offsets'
        split = self.outputType.currentText() == 'Per footprint length'
        operation = str(self.mmBox.currentText())

        transcriptFilename = str(self.fileNameInput.currentText()).replace(".bam","")
        outputDirectory = str(self.directoryOutput.text())

        print ("using lengths:", lengths)

        return useOffset, split, operation, lengths, relative, transcriptFilename, dataFiles, outputDirectory

    def browseTranscriptFileName(self):
        filename= QFileDialog.getOpenFileName(self, 'File', '.', '*.gtf *.bed')
        self.fileNameInput.addItem(filename)

    def browseFileName(self):
        filename= QFileDialog.getOpenFileName(self, 'File', '.', '*.bam')
        self.addFile(filename)

    def browseDirectory(self):

        dirname = QFileDialog.getExistingDirectory(self, 'File', '.')

        self.directoryOutput.setText(dirname)




    def addFile(self, filename):
        if filename == '':
            return
        item = QTreeWidgetItem()
        item.setData(0, Qt.DisplayRole, filename)
        item.setData(1, Qt.CheckStateRole,  Qt.Checked )

        self.fileNameView.addTopLevelItem(item)
        self.fileNames.append(str(filename))

class introPage (QWizardPage):
    def __init__(self, parent):
        super(introPage, self).__init__(parent)
        self.setTitle("Introduction")
        self.layout = QVBoxLayout()
        self.label = QLabel("This wizard will guide you through the process\nof setting up Shoelaces.\n\nYou will need an indexed sequence alignment data file\nin BAM format and gene annotations file in GTF format.")
        self.layout.addWidget(self.label)
        self.setLayout(self.layout)

class finishPage(QWizardPage):
    def __init__(self, parent):
        self.layout = QVBoxLayout()

class resourcePage (QWizardPage):
    def __init__(self, parent, title, description, resources, fileType):
        super(resourcePage, self).__init__(parent)
        self.setTitle(title)
        self.layout = QVBoxLayout()
        self.resourceView = QListView(self)
        self.resourceView.setAutoFillBackground(True)

        model = QStandardItemModel(self.resourceView)
        self.resourceView.setModel(model)
        self.resourceView.setItemDelegate(resourceItemDelegate(resources, self.resourceView))
        self.resourceView.setWordWrap(True)

        self.resourceView.setSelectionMode(QAbstractItemView.SingleSelection)
        self.resourceView.clicked.connect(self.resourceSelected)
        self.resourceView.setContextMenuPolicy(Qt.CustomContextMenu);

        self.desc = QLabel(description,self)
        self.openFileButton = QPushButton("Open file..", self)
        self.openFileButton.pressed.connect(self.openFile)
        self.fileType = fileType
        self.layout.addWidget(self.desc)
        self.layout.addWidget(self.openFileButton)
        self.layout.addWidget(self.resourceView)
        self.setLayout(self.layout)

        self.selectedLabel =QLineEdit();
        self.pageTitle= title

        self.registerField(title+"*", self.selectedLabel)
        self.updateView()

    def openFile(self):
        filename, _ = QFileDialog.getOpenFileName(settings.getMainApp(), "Open File", '', self.fileType)
        if not filename:
            return;
        settings.getMainApp().loadFile(filename)
        self.updateView();

    def updateView(self):
        resources = settings.getMainApp().resources.getResourcesDict()

        self.resourceView.model().clear()

        for name, resource in resources.items():
            resourceItem = QStandardItem(resource.GetName())

            if resource.GetFileEXT() in self.fileType:
                self.resourceView.model().appendRow(resourceItem)

    def resourceSelected(self, index):

        #item = self.resourceView.model().data(index, Qt.DisplayRole)
        key = str(index.data(Qt.DisplayRole))
        settings.getMainApp()
        self.selectedLabel.setText(key)

    def getResource(self):
        return str(self.selectedLabel.text())

class wizardDialog(QWizard):
    def __init__(self,parent, resources, startup):
        super(wizardDialog, self).__init__(parent)
        self.addPage(introPage(self))
        self.seqResPage = resourcePage(self,"Sequence alignment (BAM)","Select a sequence alignment file.", resources, "*.bam")
        self.addPage( self.seqResPage)
        self.transResPage =resourcePage(self,"Gene annotations (GTF)","Select a gene annotations file.", resources, "*.gtf")
        self.addPage(self.transResPage)
        self.noiseResPage =resourcePage(self,"Gene annotations (GTF)","Select a gene annotations file to define noise regions (optional)", resources, "*.gtf")
        self.addPage(self.noiseResPage)

        widge = QWidget()
        side_widge_layout = QGridLayout(widge)
        widge.setLayout(side_widge_layout)

        self.runAtStartupCheck = QCheckBox()
        self.runAtStartupCheck.setChecked(startup)
        side_widge_layout.addWidget(QLabel("Run wizard at startup"), 0, 0)
        side_widge_layout.addWidget(self.runAtStartupCheck,0,1)

        self.setSideWidget(widge)
        self.currentIdChanged.connect(self.checkStatus)
        self.calcOffset = False;
        layout = [QWizard.BackButton,QWizard.CancelButton, QWizard.NextButton]
        self.setButtonLayout(layout)
        self.setOption(QWizard.NoBackButtonOnStartPage,True)

    def getResources(self):
        res = {}
        res['sequence'] = self.seqResPage.getResource()
        res['transcript'] = self.transResPage.getResource()
        res['noise'] = self.noiseResPage.getResource()

        return res;

    def runAtStartup(self):
        return self.runAtStartupCheck.isChecked()

    def checkStatus(self, id):
        if (self.page(id) == self.noiseResPage):
            self.setOption(QWizard.HaveCustomButton1, True)
            self.setButtonText(QWizard.CustomButton1, "Set offset")
            layout = [QWizard.BackButton,QWizard.CancelButton, QWizard.FinishButton, QWizard.CustomButton1]
            self.customButtonClicked.connect(self.setOffsetClicked)
            self.setButtonLayout(layout)
        else:

            layout = [QWizard.BackButton,QWizard.CancelButton, QWizard.NextButton]
            self.setButtonLayout(layout)


    def setOffsetClicked(self):
        self.calcOffset = True;
        self.done(1)

class saveButton(QPushButton):
    saveAs=pyqtSignal()

    def __init__(self,  icon, parent):
        super(saveButton,self).__init__(icon,"", parent)

    def mousePressEvent(self,e):
        if e.button()==1:
            super(saveButton,self).mousePressEvent(e)
        elif e.button()==2:

            saveAs = QAction('Save as..')
            saveAs.triggered.connect(self.saveAs)

            QMenu.exec([saveAs],e.globalPos())


############################################
## Render Widgets


class glPlotWidget(QOpenGLWidget):

    offsetUpdated = pyqtSignal(int, int)
    def __init__(self, parent=None, ID = -1, data = {}):
        super(glPlotWidget, self).__init__(parent)

        f = QSurfaceFormat()
        f.setVersion(2,1)
        f.setProfile(QSurfaceFormat.CoreProfile)

        f.setSamples(2)
        self.setFormat(f)

        self.setMinimumSize(200, 100)

        self.processor = None

        self.viewCenter = 250
        self.viewTop = 300
        self.viewBottom = -11;

        self.viewLeft = 0
        self.viewRight = 500

        self.interaction = ""
        self.updateHist = False
        self.offset = 0

        self.temp = 0
        self.ID = ID
        self.codon = "start"

        self.mode = "def"
        self.hist_array = []
        self.color_array = []
        self.maxValue = 0
        self.barSize = 0.9
        self.xAxisIncr = 1
        self.displayOffset = True;
        self.displayMaxValue = True;
        self.renderYaxis = False
        self.renderXaxis = True
        self.yAxisWidth = 1
        self.yAxis = []
        self.xAxis = []


        self.textOverlayImage = QImage()
        self.textTexture = -1

        self.imageSize = QSize(1, 1)

        self.originalHistogram = {}
        self.glElementFormat = GL_QUADS
        self.glElementSize = 4



        if data:
            self.updatePlot(data)
        self.saveGraphButton = saveButton(QIcon(settings.appDir()+"/shoelaces/resources/save-icon.png"), self)
        self.saveGraphButton.clicked.connect(self.saveGraph)
        self.saveGraphButton.saveAs.connect(self.saveGraphAs)
        self.saveGraphButton.setVisible(False)
        self.setMouseTracking(True)


    def updatePlot(self, histogram, alternatingColor = [[1.0,0,0],[0.3,0.5,0.8]]):
        return

    def updateTextOverlay(self, fontSize = 12):
        self.textOverlayImage = QImage(self.imageSize, QImage.Format_ARGB32)
        font = QFont("Helvetica", fontSize)
        metric = QFontMetrics(font)
        painter = QPainter(self.textOverlayImage)
        painter.setCompositionMode(QPainter.CompositionMode_Source)
        painter.fillRect(0,0,self.imageSize.width(),self.imageSize.height(),QColor(0,0,0,0))
        painter.setFont(font);


        yRatio = self.imageSize.height() / float((self.viewTop - self.viewBottom))
        xRatio = self.imageSize.width() / float((self.viewRight - self.viewLeft))

        if self.displayMaxValue:
            text = '{:.0f}'.format(self.maxValue)
            painter.drawText(self.imageSize.width()- metric.width(text), 10*yRatio, text)

        ###########################
        ## X axis
        if self.renderXaxis:
            for (v1,v2) in self.xAxis:
                value = v1[0]
                text =  str(round(value))
                pos = self.glToImagePos(v1)
                width = metric.width(text)/2
                yPos =self.imageSize.height() - (5*yRatio)
                # if self.imageSize.height()  - yPos  < metric.height():
                #
                #     yPos -=  metric.height() - (self.imageSize.height()  - yPos)
                painter.drawText(pos[0] - width , yPos, text)



        ###########################
        ## Y axis
        if self.renderYaxis:
            for (v1,v2) in self.yAxis:

                value = self.posToValue(v1[1])
                text = str(round(value))
                width = metric.width(text)
                pos = self.glToImagePos(v1)
                painter.drawText( self.imageSize.width() - width ,  pos[1], text)

        if self.displayOffset:
            text = "Offset: " + str(self.offset+self.temp)

            painter.drawText(5, 10*yRatio, text)
        painter.end()

    def createXAxis(self, left, right):

        self.xAxis = []
        if self.renderYaxis:
            right -= 0.5

        xoffset = self.barSize*0.5

        for i in range(int(left),int(right+1), self.xAxisIncr):

            self.xAxis.append(([i + xoffset, self.viewBottom, 0],[i + xoffset,self.viewBottom +5, 0]))

    def setOffset(self, offset):
        return

    def renderTranscript(self):
        return

    def render(self):
        if not self.hist_array:
            return

        if self.updateHist and self.hist_array:

            glBindBuffer (GL_ARRAY_BUFFER, self.vboHist)
            glBufferData (GL_ARRAY_BUFFER, len(self.hist_array)*self.glElementSize,  (c_float*len(self.hist_array))(*self.hist_array), GL_STATIC_DRAW)

            glBindBuffer (GL_ARRAY_BUFFER, self.cboCol)
            glBufferData (GL_ARRAY_BUFFER, len(self.color_array)*self.glElementSize,  (c_float*len(self.color_array))(*self.color_array), GL_STATIC_DRAW)



        self.renderTranscript()

        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()

        drawLeft = self.viewLeft + (self.offset+self.temp)
        drawRight = self.viewRight + (self.offset+self.temp)


        gluOrtho2D(drawLeft, drawRight, self.viewBottom, self.viewTop)
        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()

        glBindBuffer(GL_ARRAY_BUFFER, self.cboCol)
        glColorPointer(3,GL_FLOAT,0,None)


        glBindBuffer (GL_ARRAY_BUFFER, self.vboHist)
        glVertexPointer(3, GL_FLOAT, 0, None)

        glEnableClientState(GL_VERTEX_ARRAY)
        glEnableClientState(GL_COLOR_ARRAY)
        glDrawArrays(self.glElementFormat, 0, int(len(self.hist_array)/3))

    def renderText(self):
        if not self.hist_array:
            return
        if not self.textTexture:
            glGenTextures( 1, self.textTexture );
            glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, self.textOverlayImage.width(), self.textOverlayImage.height(), 0, GL_RGBA, GL_UNSIGNED_BYTE, None )


        glEnable(GL_BLEND)
        glEnable(GL_TEXTURE_2D)

        ptr = self.textOverlayImage.bits()
        ptr.setsize(self.textOverlayImage.byteCount())

        ## copy the data out as a string
        strData = ptr.asstring()

        ## get a read-only buffer to access the data
        buf = memoryview(ptr)

        ## view the data as a read-only numpy array

        arr = np.frombuffer(buf, dtype=np.ubyte).reshape(self.textOverlayImage.height(), self.textOverlayImage.width(), 4)

        glActiveTexture(GL_TEXTURE0)

        glBindTexture( GL_TEXTURE_2D, self.textTexture );

        glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
        glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, self.textOverlayImage.width(), self.textOverlayImage.height(), 0, GL_RGBA, GL_UNSIGNED_BYTE, arr )

        glMatrixMode(GL_PROJECTION);
        glLoadIdentity();
        gluOrtho2D(self.viewLeft , self.viewRight , self.viewBottom, self.viewTop)
        glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();
        #glTranslate(0,self.viewBottom+5,0)
        glBegin(GL_QUADS)

        glTexCoord2f(0.0,1.0)
        glVertex3f(self.viewLeft,self.viewBottom,0.0)
        glTexCoord2f(1.0,1.0)
        glVertex3f(self.viewRight,self.viewBottom,0.0)
        glTexCoord2f(1.0,0.0)
        glVertex3f(self.viewRight,self.viewTop,0.0)
        glTexCoord2f(0.0,0.0)
        glVertex3f(self.viewLeft,self.viewTop,0.0)

        glEnd()
        glBindTexture(GL_TEXTURE_2D, 0)

    def renderScale(self):
        if not self.hist_array:
            return
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()

        gluOrtho2D(self.viewLeft , self.viewRight , self.viewBottom, self.viewTop)

        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()

        glColor3f(0,0,0)

        glLineWidth(1)
        glLineStipple(1,0x0307)
        glEnable(GL_LINE_STIPPLE)

        glBegin(GL_LINES)
        glVertex3f(self.viewLeft,100 ,0)
        glVertex3f(self.viewRight,100 ,0)
        glEnd()
        glDisable(GL_LINE_STIPPLE)

    def renderaxis(self, l, r, t, b):
        if not self.hist_array:
            return

        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()

        gluOrtho2D(self.viewLeft , self.viewRight , self.viewBottom, self.viewTop)
        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()

        glBegin(GL_LINES)
        if self.renderXaxis:
            for pos in self.xAxis:
                glVertex(pos[0])
                glVertex(pos[1])

        if self.renderYaxis:
            for pos in self.yAxis:
                glVertex(pos[0])
                glVertex(pos[1])

        glEnd()

    def paintGL(self):
        glClearColor(1.0,1.0,1.0,1.0)
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)

        self.render()
        self.renderScale()
        self.renderaxis(self.viewLeft,self.viewRight, self.viewTop, self.viewBottom)

        self.renderText()

        glFlush()

    def resizeGL(self, w, h):

        glViewport(0, 0, w, h)
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()

        self.imageSize = QSize(w,h)

        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, self.imageSize.width(), self.imageSize.height(), 0, GL_RGBA, GL_UNSIGNED_BYTE,None )

        self.updateTextOverlay()

    def initializeGL(self):


        self.vboHist = glGenBuffers(1)


        self.cboCol = glGenBuffers(1)
        glutInit()

        glClearColor(0.0, 0.0, 0.0, 1.0)
        glClearDepth(1.0)

        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        #gluPerspective(40.0, 1.0, 1.0, 30.0)

        self.updateTextOverlay()

    def saveGraphAs(self):
        filename, _ = QFileDialog.getSaveFileName(self, "Save Image", '',
                    "images (*.png)")
        if not filename:
            return

        self.saveGraph(filename)

    def saveGraph(self, filename=""):
        if filename == "" or type(filename) is bool:
            filename=str(datetime.datetime.now())+'.png'
        s = self.size()
        self.resize(640,480)
        self.updateTextOverlay(24)
        self.update()
        output = self.grabFramebuffer()
        self.resize(s)
        output.save(filename)

    def getCanvasPos(self, pos):

        relX = pos.x() / float(self.width())
        relY = pos.y() / float(self.height())
        viewX = (relX* float(self.viewRight - self.viewLeft)) + self.viewLeft
        viewY = (relY* float(self.viewTop - self.viewBottom)) + self.viewBottom
        result = QtCore.QPoint(viewX, viewY)
        return result

    def eventFilter(self, widget, event):

        return super(self).eventFilter(self, widget, event)

    def enterEvent(self, QEvent):
        self.saveGraphButton.setVisible(True)

    def leaveEvent(self, QEvent):
        self.saveGraphButton.setVisible(False)

    def glToImagePos(self,pos):
        xRatio = self.imageSize.width() / float((self.viewRight - self.viewLeft))
        x = pos[0]

        x -= self.viewLeft
        x *= xRatio

        y = pos[1]
        y -= self.viewBottom
        y /= float(self.viewTop - self.viewBottom)
        y = 1 - y
        y *= self.imageSize.height()

        return (x , y )

    def imageToGlPos(self,pos):
        x = float(pos.x()) / float(self.imageSize.width());
        y = 1.0 - (float(pos.y()) / float(self.imageSize.height()));



        x = x * (self.viewRight - self.viewLeft) + self.viewLeft;
        y = y * (self.viewTop - self.viewBottom) + self.viewBottom;



        return (x , y )

    def posToValue(self, y):
        if y < 0:
            return 0
        y /= (self.viewTop-10)
        if y >= 1.0:
            return self.maxValue
        return y * self.maxValue

    def valueToPos(self, value):
        y = value / self.maxValue
        y *= self.viewTop - 10
        return y

    # def mouseMoveEvent(self, e):
    #     tempPos =e.pos()
    #     relY = tempPos.y() / float(self.height())
    #     relY = 1 - relY
    #
    #     viewY = (relY* float(self.viewTop - self.viewBottom)) + self.viewBottom
    #     print(self.posToValue(viewY))

class glStatsPlotWidget(glPlotWidget):

    offsetUpdated = pyqtSignal(int, int)

    def __init__(self, parent=None, ID = -1, histogram = {}, codon = "start"):
        super(glStatsPlotWidget, self).__init__(parent, ID, histogram)

        self.codon = codon
        self.mode = "hist"
        self.renderYaxis = True
        self.displayOffset = False

    def updatePlot(self, histogram, alternatingColor = [[1.0,0,0],[0.3,0.5,0.8]]):
        glDeleteBuffers(2, [self.vboHist, self.cboCol])
        self.vboHist = glGenBuffers(1)
        self.cboCol = glGenBuffers(1)

        self.hist_array = []
        self.color_array = []
        maxValue = 1;
        maxX = 0
        minX = 0

        counter = 0

        for diff in sorted(histogram.keys()):

            value = histogram[diff]

            color = alternatingColor[1]
            self.color_array += color
            counter +=1;

            self.hist_array += [float(diff), float(value), 0]
            self.hist_array += [float(diff), 0, 0]
            self.hist_array += [float(diff) + self.barSize, 0, 0]
            self.hist_array += [float(diff) + self.barSize, float(value), 0]

            self.color_array += color
            self.color_array += color
            self.color_array += color

            maxValue = max(value, maxValue)

            maxX = max(diff, maxX)
            minX = min(diff, minX)


        self.maxValue = maxValue

        if self.maxValue > 0:
            fScale = 100.0/float(self.maxValue)
        else:
            fScale = 0
        for i in range(0, len(self.hist_array), 3):
            self.hist_array[i+1] *= fScale

        self.viewTop = 100 + 10


        self.viewLeft = float(min(histogram.keys()))


        self.viewRight = float(max(histogram.keys())+self.barSize)
        if self.renderYaxis:
            self.viewRight += 0.3
            self.yAxisWidth = 0.1

        self.updateHist = True
        self.updateTextOverlay()
        self.update()

        self.originalHistogram = histogram

        self.createXAxis(self.viewLeft,self.viewRight)
        self.createYAxis(0, self.viewTop)

    def createYAxis(self, bottom, top):
        self.yAxis = []
        if not self.hist_array or not self.maxValue:
            return

        numberOfDigits = int(math.log10(self.maxValue))
        divider = math.pow(10,numberOfDigits)



        numberOfLines = self.maxValue/divider
        while (numberOfLines < 2 and divider >=2):
            divider /= 2.0
            numberOfLines = self.maxValue/divider




        for i in range(0,self.maxValue, int(divider)):
            ypos = self.valueToPos(i)
            self.yAxis.append(([self.viewRight, ypos, 0],[self.viewRight - self.yAxisWidth ,ypos, 0] ))

class glGenePlotWidget(glPlotWidget):
    def __init__(self, parent, transcript = None, reads = []):


        super(glGenePlotWidget, self).__init__( parent)
        self.setMinimumSize(500, 300)
        self.transcripts = []
        self.noncoding= []

        self.reads = reads
        self.transcripThickness = 10
        self.maxWidth = 500
        self.padding  = 100
        self.gap = self.padding/10
        self.vgap = self.padding/20

        self.vboHist = 0
        self.exonBins  = []
        self.renderYaxis = False
        self.renderXaxis = False
        self.displayOffset = False
        self.binTorender = None

    def checkOverlap(self, pos, size):
        res = []
        for  bin in self.exonBins:
            if pos < bin['start'] and pos + size > bin['start']:
                res.append( bin)
            if pos >= bin['start'] and pos <=  bin['end']:
                res.append(bin)
            if pos < bin['start'] and pos+size > bin['end']:
                res.append( bin)
        return res

    def updateGene(self, gene):

        self.noncoding = []
        self.transcripts = []
        for key, t in gene['transcripts'].items():

            self.transcripts.append(t)

        self.noncoding = []
        for key, t in gene['non_coding'].items():
            self.noncoding.append(t)

        self.updateTranscriptData()

    def addTranscriptToBins(self, transcript):
        for size, pos in zip(transcript['exon_sizes'], transcript['exon_positions']):
            currentBins = self.checkOverlap(pos, size)



            if currentBins:
                minStart = currentBins[0]['start']
                maxEnd = currentBins[0]['end']

                for b in currentBins:
                    minStart = min(b['start'],minStart)
                    maxEnd = max(b['end'], maxEnd)
                    if b in self.exonBins:
                        self.exonBins.remove(b)
                bin = {}
                bin['start'] = min(minStart,pos)
                bin['end'] = max(maxEnd, pos + size)
                bin['size'] = bin['end'] - bin['start']
                self.exonBins.append(bin)

            else:
                bin = {}
                bin['start'] = pos
                bin['end'] = pos + size
                bin['size'] = size
                self.exonBins.append(bin)

    def updateTranscript(self, transcripts):
        self.noncoding = []
        self.transcripts = transcripts.copy()

        self.updateTranscriptData()

    def updateTranscriptData(self):
        self.exonBins = []
        reverseStrand = False;
        for t in self.transcripts:
            self.addTranscriptToBins(t)
            #reverseStrand = not t['forward']
        for t in self.noncoding:
            self.addTranscriptToBins(t)

        self.exonBins = sorted(self.exonBins, key=lambda bin: bin['start'] )


        self.gap= self.padding/(len(self.exonBins) + 2)
        self.transcriptWidth  = 0
        currentPos = 0;

        for i, bin in enumerate(self.exonBins):

            self.transcriptWidth += bin['size'] + self.gap
            bin['draw_start'] = currentPos
            bin['draw_end'] = currentPos + bin['size'] + 1
            print(bin['start'], currentPos)
            currentPos = bin['draw_end'] +  self.gap







        self.maxWidth = self.padding + self.transcriptWidth
        self.viewLeft = 0
        self.viewCenter = self.maxWidth / 2
        self.viewRight = self.maxWidth
        if reverseStrand:
            self.viewLeft = self.maxWidth
            self.viewRight = 0

        self.transcripThickness = self.vgap - 2

        self.viewBottom = - (len(self.transcripts)+len(self.noncoding)) * (self.transcripThickness*2 + self.vgap)

        self.update()



    def updateData(self, histogram):
        glDeleteBuffers(3, [self.vboHist, self.cboCol, self.cboCol2])
        self.vboHist = glGenBuffers(1)
        self.cboCol = glGenBuffers(1)
        self.cboCol2 = glGenBuffers(1)
        self.reads_array = []


        self.hist_array = []
        self.color_array = []
        maxValue = 0;

        histogramPos = sorted(histogram)


        color_array_2 = []
        for pos in histogramPos:
            value = histogram[pos]

            bin  = self.checkOverlap(pos, 1)
            if not bin:
                continue
            diff = pos - bin[0]['start']
            x = bin[0]['draw_start'] + diff

            self.hist_array += [x, value, 0]
            self.hist_array += [x, 0, 0]
            if pos % 3 == 1:
                color= [1.0,0.0,0.0]
            else:
                color = [0.3,0.5,0.8]
            self.color_array += color
            self.color_array += color

            color_array_2 += [0.3,0.5,0.8]
            color_array_2 += [0.3,0.5,0.8]

            maxValue = max(value, maxValue)



        self.maxValue = maxValue
        if self.maxValue > 0:
            fScale = 100.0/float(self.maxValue)
        else:
            fScale = 0
        for i in range(0, len(self.hist_array), 3):
            self.hist_array[i+1] *= fScale




        self.viewTop = 100 + 10


        glBindBuffer (GL_ARRAY_BUFFER, self.vboHist)

        glBufferData (GL_ARRAY_BUFFER, len(self.hist_array)*4,  (c_float*len(self.hist_array))(*self.hist_array), GL_STATIC_DRAW)

        glBindBuffer (GL_ARRAY_BUFFER, self.cboCol)
        glBufferData (GL_ARRAY_BUFFER, len(self.color_array)*4,  (c_float*len(self.color_array))(*self.color_array), GL_STATIC_DRAW)

        glBindBuffer (GL_ARRAY_BUFFER, self.cboCol2)
        glBufferData (GL_ARRAY_BUFFER, len(self.color_array)*4,  (c_float*len(color_array_2))(*color_array_2), GL_STATIC_DRAW)
        self.updateTextOverlay()
    def initializeGL(self):
        self.vboHist = glGenBuffers(1)
        self.cboCol = glGenBuffers(1)
        self.cboCol2 = glGenBuffers(1)
        glutInit()
        glClearColor(0.0, 0.0, 0.0, 1.0)
        glClearDepth(1.0)
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        self.updateTextOverlay()

    def paintGL(self):

        glClearColor(1.0,1.0,1.0,1.0)
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)

        i = 0
        for t in self.transcripts:
            self.renderTranscript(t, i)
            i+=1

        for t in self.noncoding:
            self.renderTranscript(t, i, [0.8,0.1,0.1])
            i+=1

        if not self.transcripts or not self.vboHist:
            return

        glOrtho(self.viewLeft, self.viewRight, self.viewBottom, self.viewTop,-1,1)


        color = [0.1,0.1,0.8]
        glColor3fv(color)
        windowWidth = self.width()
        viewWidth = abs(self.viewRight - self.viewLeft)
        pixelsPerNuc = float(windowWidth)/float(viewWidth)

        pixelRatio = settings.getMainApp().devicePixelRatio()
        self.lineWidth = max(3,pixelsPerNuc*0.9*pixelRatio)

        maxWidth = glGetFloat(GL_ALIASED_LINE_WIDTH_RANGE)
        self.lineWidth = min(self.lineWidth, maxWidth[1])

        glLineWidth(self.lineWidth)
        lineImprint = self.lineWidth / float(self.imageSize.width())
        lineImprint *= viewWidth

        #print(maxWidth, self.lineWidth)
        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()

        glColor3f(0.3, 0.5, 0.8)
        glTranslatef(lineImprint/2.0,0,0)

        if self.lineWidth <4:
            glBindBuffer(GL_ARRAY_BUFFER, self.cboCol2)
        else:
            glBindBuffer(GL_ARRAY_BUFFER, self.cboCol)

        glColorPointer(3,GL_FLOAT,0,None)
        glBindBuffer (GL_ARRAY_BUFFER, self.vboHist)
        glVertexPointer(3, GL_FLOAT, 0, None)
        glEnableClientState(GL_VERTEX_ARRAY)
        glEnableClientState(GL_COLOR_ARRAY)
        glDrawArrays(GL_LINES, 0, len(self.hist_array))

        glColor3f(0,0,0)
        glLineWidth(2)
        self.renderText()
        glLineStipple(1,0x0307)
        glEnable(GL_LINE_STIPPLE)

        glBegin(GL_LINES)
        glVertex3f(self.viewLeft,100 ,0)
        glVertex3f(self.viewRight,100,0)
        glEnd()
        glDisable(GL_LINE_STIPPLE)
        #self.renderBin(self.binTorender)
        glFlush()


    def renderTranscript(self, transcript, idx, color = [0.1,0.1,0.8]):

        if not transcript:
            return

        reverseStrand = not transcript['forward']

        exonSizes = transcript['exon_sizes']
        exonPositons = transcript['exon_positions']

        #currentPos = self.gap


        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        gluOrtho2D(self.viewLeft, self.viewRight, self.viewBottom, self.viewTop)
        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()

        glColor3fv(color)
        glEnable(GL_DEPTH_TEST)
        glBegin(GL_QUADS)

        thickness = self.transcripThickness
        center =  -(self.vgap * idx + self.transcripThickness *2 * idx) - self.vgap

        for i, s in enumerate(exonSizes):
            glColor3fv(color)
            bin = self.checkOverlap(exonPositons[i], s)
            #endBin = self.checkOverlap(exonPositons[i]+s, 1)

            size_render = s + 1
            if not bin:
                continue
            # if(endBin != bin and endBin):
            #
            #     size_render -= (endBin['start'] - bin['end'])
            #     size_render += self.gap

            diff = exonPositons[i] - bin[0]['start']

            currentPos = bin[0]['draw_start'] + diff


            if (i == transcript['start_exon'] and not reverseStrand) or (i == transcript['end_exon'] and  reverseStrand):


                toCDS = transcript['CDS_min'] - exonPositons[i]



                glVertex(currentPos, center - thickness, 0)
                glVertex(currentPos + toCDS, center - thickness, 0)
                glVertex(currentPos + toCDS, center + thickness, 0)
                glVertex(currentPos, center + thickness, 0)

                thickness *= 1.2

                glVertex(currentPos + toCDS, center - thickness, 0)
                glVertex(currentPos + size_render, center  - thickness, 0)
                glVertex(currentPos + size_render, center + thickness, 0)
                glVertex(currentPos + toCDS, center + thickness, 0)


                if reverseStrand:
                    glColor3fv([0.8,0.1,0.1]);

                    glVertex(currentPos + toCDS-3, center - thickness, 1)
                    glVertex(currentPos+ toCDS, center - thickness, 1)
                    glVertex(currentPos + toCDS, center + thickness, 1)
                    glVertex(currentPos + toCDS-3, center + thickness, 1)
                else:


                    glColor3fv([0.1,0.8,0.1]);

                    glVertex(currentPos+ toCDS, center - thickness, 1)
                    glVertex(currentPos+ toCDS + 3, center - thickness, 1)
                    glVertex(currentPos+ toCDS + 3, center + thickness, 1)
                    glVertex(currentPos+ toCDS, center + thickness, 1)


            elif (i == transcript['end_exon'] and not reverseStrand) or (i == transcript['start_exon'] and  reverseStrand):

                toCDS = transcript['CDS_max'] - exonPositons[i] + 1

                if  reverseStrand:
                    glColor3fv([0.1,0.8,0.1]);

                    glVertex(currentPos + toCDS-3, center - thickness, 1)
                    glVertex(currentPos + toCDS, center - thickness, 1)
                    glVertex(currentPos + toCDS, center + thickness, 1)
                    glVertex(currentPos + toCDS-3, center + thickness, 1)
                else:


                    glColor3fv([0.8,0.1,0.1]);

                    glVertex(currentPos+ toCDS, center - thickness, 1)
                    glVertex(currentPos+ toCDS + 3, center - thickness, 1)
                    glVertex(currentPos+ toCDS + 3, center + thickness, 1)
                    glVertex(currentPos+ toCDS, center + thickness, 1)
                glColor3fv(color)
                glVertex(currentPos, center - thickness, 0)
                glVertex(currentPos + toCDS, center - thickness, 0)
                glVertex(currentPos + toCDS, center + thickness, 0)
                glVertex(currentPos, center + thickness, 0)

                thickness /= 1.2

                glVertex(currentPos + toCDS, center - thickness, 0)
                glVertex(currentPos + size_render, center - thickness, 0)
                glVertex(currentPos + size_render, center + thickness, 0)
                glVertex(currentPos + toCDS , center + thickness, 0)


            else:
                glVertex(currentPos, center - thickness, 0)
                glVertex(currentPos + size_render, center - thickness, 0)
                glVertex(currentPos + size_render, center + thickness, 0)
                glVertex(currentPos, center + thickness, 0)




        glEnd()

    def renderBin(self, bin):

        if not bin:
            return
        print(bin)
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        gluOrtho2D(self.viewLeft, self.viewRight, self.viewBottom, self.viewTop)
        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()

        glColor3fv([0.1,0.5,0.1])
        glDisable(GL_DEPTH_TEST);
        glBegin(GL_QUADS)

        center =  0;
        thickness = 10;


        glVertex(bin['draw_start'], center - thickness, 0)
        glVertex(bin['draw_end'] , center - thickness, 0)
        glVertex(bin['draw_end'] , center + thickness, 0)
        glVertex(bin['draw_start'] , center+ thickness, 0)


        glEnd()

    def mousePressEvent(self, e):



        if e.button() == 1:
            self.interaction = "translate"
        elif e.button() == 2:
            self.interaction = "zoom"

        canvasPos = self.getCanvasPos(e.pos())

        self.startPos = canvasPos

        self.newCenter = canvasPos.x()
        self.update()

    def mouseMoveEvent(self, e):

        self.currentPos = self.getCanvasPos(e.pos())
        (x,y) = self.imageToGlPos(e.pos());

        self.binTorender = self.getBin(x)
        if self.binTorender:
            ref_pos = x - self.binTorender['draw_start'] + self.binTorender['start']
            print(round(ref_pos), round(self.posToValue(y) + 0.5))

        if self.interaction == "zoom":

            diff = self.currentPos.x() - self.startPos.x()


            currentWidth =  abs(self.viewRight - self.viewLeft)
            ratioLeft = abs(self.newCenter - self.viewLeft) / float(currentWidth)
            ratioRigth = abs(self.viewRight - self.newCenter) / float(currentWidth)

            if  abs(self.viewRight - self.viewLeft) > diff:
                self.viewLeft += diff * ratioLeft
                self.viewRight -= diff * ratioRigth
                self.viewCenter = (self.viewRight + self.viewLeft)/2.0



        elif self.interaction == "translate":

            diff = self.currentPos.x() - self.startPos.x()
            self.viewLeft -= diff
            self.viewRight -= diff

            self.viewCenter = (self.viewRight + self.viewLeft)/2

        self.update()

    def mouseReleaseEvent(self, QMouseEvent):

        self.interaction = ""

    def getBin(self, glPos_x):
        for bin in self.exonBins:
            if bin['draw_start'] <= glPos_x <= bin['draw_end']:
                return bin

        return None

class glOffsetPlotWidget(glPlotWidget):

    offsetUpdated = pyqtSignal(int, int)

    def __init__(self, parent=None, ID = -1, histogram = {}, codon = "start"):
        super(glOffsetPlotWidget, self).__init__(parent, ID, histogram)

        self.codon = codon
        self.mode = "hist"
        self.xAxisIncr = 5

    def updatePlot(self, histogram, alternatingColor = [[1.0,0,0],[0.3,0.5,0.8]]):

        self.hist_array = []
        self.color_array = []
        maxValue = 1;
        maxX = 0
        minX = 0
        counter = 0;
        for diff in sorted(histogram.keys()):

            value = histogram[diff]
            colorOffset = ((self.temp+self.offset-diff)%3)
            if self.codon == 'stop':
                colorOffset = ((self.temp+self.offset-diff)%3)

            if colorOffset % 3 == 0:
                color= alternatingColor[0]
            else:
                color = alternatingColor[1]
            self.color_array += color
            counter +=1;

            self.hist_array += [float(diff), float(value), 0]
            self.hist_array += [float(diff), 0, 0]
            self.hist_array += [float(diff) + self.barSize, 0, 0]
            self.hist_array += [float(diff) + self.barSize, float(value), 0]

            self.color_array += color
            self.color_array += color
            self.color_array += color

            maxValue = max(value, maxValue)

            maxX = max(diff, maxX)
            minX = min(diff, minX)

        self.maxValue = maxValue

        if self.maxValue > 0:
            fScale = 100.0/float(self.maxValue)
        else:
            fScale = 0
        for i in range(0, len(self.hist_array), 3):
            self.hist_array[i+1] *= fScale

        self.viewTop = 100 + 10

        if self.codon == "start":
            self.viewLeft = float(-30)
            self.viewRight = float(10)
        else:
            self.viewLeft = float(-10)
            self.viewRight = float(30)
        self.updateHist = True
        self.updateTextOverlay()
        self.update()

        self.originalHistogram = histogram

        self.createXAxis(self.viewLeft,self.viewRight)

    def setOffset(self, offset):
        self.offset = offset

        if self.mode == "def":
            self.updateHist = True
        else:
            self.updatePlot(self.originalHistogram)
        self.updateTextOverlay()

        self.update()

    def renderTranscriptPart(self, thickness, start, end):
        glBegin(GL_QUADS)

        center = (self.viewTop + self.viewBottom) / 2.0

        glVertex(start, center - thickness, 0)
        glVertex(end, center - thickness, 0)
        glVertex(end, center + thickness, 0)
        glVertex(start, center + thickness, 0)

        glEnd()

    def renderTranscript(self):

        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()

        gluOrtho2D(self.viewLeft, self.viewRight , self.viewBottom, self.viewTop)

        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()

        glColor4f(0.1,0.1,0.8,0.3)

        thin = (self.viewTop + self.viewBottom) / 4.0
        thick = thin * 1.6
        if self.codon == "start":
            self.renderTranscriptPart(thin, self.viewLeft, 0)
            self.renderTranscriptPart(thick,3,self.viewRight)
            glColor4f(0.1,0.8,0.1,0.4)
            self.renderTranscriptPart(thick,0,3)
        else:
            self.renderTranscriptPart(thick, self.viewLeft, 0)
            self.renderTranscriptPart(thin,3,self.viewRight)
            glColor4f(0.8,0.1,0.1,0.4)
            self.renderTranscriptPart(thick,0,3)

    def mousePressEvent(self, e):

        if e.button() == 1:
            self.interaction = "translate"

        canvasPos = self.getCanvasPos(e.pos())

        self.startPos = canvasPos


        self.update()

    def mouseMoveEvent(self, e):




        self.currentPos = self.getCanvasPos(e.pos())

        if self.interaction == "translate":


            diff = self.currentPos.x() - self.startPos.x()
            self.temp = -int(diff)
            if self.mode == "def":
                    self.updateHist = True
            else:
                self.updatePlot(self.originalHistogram)




        self.updateTextOverlay()
        self.update()

    def mouseReleaseEvent(self, QMouseEvent):
        self.offset += self.temp
        self.temp = 0
        self.offsetUpdated.emit( self.ID,self.offset)
        self.interaction = ""

class glGraphPlotWidget(glPlotWidget):

    offsetUpdated = pyqtSignal(int, int)

    def __init__(self, parent=None, ID = -1, histogram = {}, codon = "start"):
        super(glGraphPlotWidget, self).__init__(parent, ID, {})

        self.glElementType = GL_LINE_STRIP
        self.glElementSize = 1

        self.displayOffset = False
        self.barSize = 0.1
        self.codon = codon
        self.mode = "graph"
        self.updatePlot(histogram)




    def updatePlot(self, histogram, alternatingColor = [[1.0,0,0],[0.3,0.5,0.8]]):
        self.hist_array = []
        self.color_array = []
        maxValue = 1;
        maxX = 0
        minX = 0

        counter = 0;

        for diff in sorted(histogram.keys()):
            value = histogram[diff]
            color = alternatingColor[1]
            self.color_array += color
            counter +=1;
            self.hist_array += [float(diff), float(value), 0]
            maxValue = max(value, maxValue)

            maxX = max(diff, maxX)
            minX = min(diff, minX)


        self.maxValue = maxValue

        if self.maxValue > 0:
            fScale = 100.0/float(self.maxValue)
        else:
            fScale = 0
        for i in range(0, len(self.hist_array), 3):
            self.hist_array[i+1] *= fScale

        self.viewTop = 100 + 10


        self.viewLeft = float(0)
        self.viewRight = float(10)

        self.updateHist = True
        self.updateTextOverlay()
        self.update()

        self.originalHistogram = histogram

        self.createXAxis(self.viewLeft,self.viewRight)

    def render(self):

        if not self.hist_array:
            return

        if self.updateHist:
            glEnableClientState (GL_VERTEX_ARRAY)
            glBindBuffer (GL_ARRAY_BUFFER, self.vboHist)
            glBufferData (GL_ARRAY_BUFFER, len(self.hist_array)*4,  (c_float*len(self.hist_array))(*self.hist_array), GL_STATIC_DRAW)


        glPointSize(10)
        glLineWidth(2)
        glColor3f(0.3,0.5,0.8)
        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        drawLeft = self.viewLeft + (self.offset+self.temp)
        drawRight = self.viewRight + (self.offset+self.temp)
        gluOrtho2D(drawLeft, drawRight, self.viewBottom, self.viewTop)
        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()

        glBindBuffer (GL_ARRAY_BUFFER, self.vboHist)
        glVertexPointer(3, GL_FLOAT, 0, None)

        glDrawArrays(GL_LINE_STRIP, 0, int(len(self.hist_array)/3))

        #self.renderXaxis(drawLeft,drawRight)

class glModularPlotWidget(glPlotWidget):

    offsetUpdated = pyqtSignal(int, int)

    def __init__(self, parent=None, ID = -1, histogram = {}, codon = "start"):
        super(glModularPlotWidget, self).__init__(parent, ID, histogram)

        self.codon = codon
        self.mode = "def"

    def updatePlot(self, histogram, alternatingColor = [[1.0,0,0],[0.3,0.5,0.8]]):

        self.hist_array = []
        self.color_array = []
        maxValue = 1;
        maxX = 0
        minX = 0

        counter = 0;

        for diff in sorted(histogram.keys()):

            value = histogram[diff]
            #colorOffset = ((self.temp+self.offset)%3)

            if (diff) % 3 == 1:
                color= alternatingColor[0]
            else:
                color = alternatingColor[1]

            counter +=1;

            self.hist_array += [float(diff), float(value), 0]
            self.hist_array += [float(diff), 0, 0]
            self.hist_array += [float(diff) + self.barSize, 0, 0]
            self.hist_array += [float(diff) + self.barSize, float(value), 0]

            self.color_array += color
            self.color_array += color
            self.color_array += color
            self.color_array += color


            maxValue = max(value, maxValue)

            maxX = max(diff, maxX)
            minX = min(diff, minX)


        self.maxValue = maxValue

        if self.maxValue > 0:
            fScale = 100.0/float(self.maxValue)
        else:
            fScale = 0
        for i in range(0, len(self.hist_array), 3):
            self.hist_array[i+1] *= fScale

        self.viewTop = 100 + 10

        self.viewLeft = float(1)
        self.viewRight = float(3+self.barSize)

        self.updateHist = True
        self.updateTextOverlay()
        self.update()

        self.originalHistogram = histogram
        self.createXAxis(self.viewLeft,self.viewRight)

    def setOffset(self, offset):
        self.offset = offset

        if self.mode == "def":
            self.updateHist = True
        else:
            self.updatePlot(self.originalHistogram)
        self.updateTextOverlay()

        self.update()

    def render(self):
        if not self.hist_array:
            return

        if self.updateHist:

            hist_arry_shifted = []
            shift = ((self.offset+self.temp) % 3)*12


            for i in range(0,len(self.hist_array),12):

                currentIndex = (i+shift) % len(self.hist_array)

                hist_arry_shifted += [self.hist_array[i], self.hist_array[currentIndex+1], self.hist_array[currentIndex+2]]
                hist_arry_shifted += [self.hist_array[i+3], self.hist_array[currentIndex+4], self.hist_array[currentIndex+5] ]
                hist_arry_shifted += [self.hist_array[i+6], self.hist_array[currentIndex+7], self.hist_array[currentIndex+8]]
                hist_arry_shifted += [self.hist_array[i+9], self.hist_array[currentIndex+10], self.hist_array[currentIndex+11]]


            glBindBuffer (GL_ARRAY_BUFFER, self.vboHist)
            glBufferData (GL_ARRAY_BUFFER, len(hist_arry_shifted)*4,  (c_float*len(hist_arry_shifted))(*hist_arry_shifted), GL_DYNAMIC_DRAW)

            glBindBuffer (GL_ARRAY_BUFFER, self.cboCol)
            glBufferData (GL_ARRAY_BUFFER, len(self.color_array)*4,  (c_float*len(self.color_array))(*self.color_array), GL_STATIC_DRAW)


        glClearColor(1.0,1.0,1.0,1.0)
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)


        glColor3f(0.3,0.5,0.8)
        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        gluOrtho2D(self.viewLeft , self.viewRight, self.viewBottom, self.viewTop)
        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()


        glBindBuffer(GL_ARRAY_BUFFER, self.cboCol)
        glColorPointer(3,GL_FLOAT,0,None)


        glBindBuffer (GL_ARRAY_BUFFER, self.vboHist)
        glVertexPointer(3, GL_FLOAT, 0, None)

        glEnableClientState(GL_VERTEX_ARRAY)
        glEnableClientState(GL_COLOR_ARRAY)
        glDrawArrays(GL_QUADS, 0, len(self.hist_array)-1)

    def mousePressEvent(self, e):

        if e.button() == 1:
            self.interaction = "translate"

        canvasPos = self.getCanvasPos(e.pos())

        self.startPos = canvasPos


        self.update()

    def mouseMoveEvent(self, e):

        self.currentPos = self.getCanvasPos(e.pos())

        if self.interaction == "translate":

            diff = self.currentPos.x() - self.startPos.x()

            self.temp = -int(diff)
            self.updateHist = True


        self.updateTextOverlay()
        self.update()

    def mouseReleaseEvent(self, QMouseEvent):
        self.offset += self.temp
        self.temp = 0
        self.offsetUpdated.emit( self.ID,self.offset)
        self.interaction = ""



###############################




import sys
import unittest
from shoelaces.mainwindow import *
#from testing import *
from PyQt5.QtWidgets import *


def cli():
    #pass

#if __name__ == '__main__':
    #unittest.main()
    if len(sys.argv) == 1:
        qApp = QApplication(sys.argv)
        qApp.setWindowIcon(QIcon(settings.appDir()+"/shoelaces/resources/logo.png"))
        aw = ApplicationWindow()
        aw.setWindowTitle("%s" % "Shoelaces")
        aw.show()
        sys.exit(qApp.exec_())


    else:
         #qApp = QCoreApplication(sys.argv)
         cla = commandLineApplication(sys.argv)


if __name__ == '__main__':
    cli()




import sys, os, random
from shoelaces.offsetwindow import *
from shoelaces.graphdockwidget import *
from shoelaces.resource import *
import argparse
import shoelaces.settings as settings
from PyQt5 import QtCore


class ApplicationWindow(QMainWindow):
    def __init__(self):

        super(ApplicationWindow, self).__init__()
        settings.init(self)
        self.setCentralWidget(None)
        self.setDockNestingEnabled(True);

        self.createMenus()
        self.createDockWidgets()
        self.createToolbar()
        self.createActions()
        self.createSignals()
        self.updateMenu()

        self.selectedtranscripts = []
        self.currentRatio = 0.1
        self.selectedGene = None
        self.geneView = None
        self.transcriptView = None
        self.activeTranscriptRes = None
        self.activeSequenceRes = None
        self.noiseResource = None
        self.configResource = ConfigResource()
        self.processor = sequenceProcessor()
        self.progress = None
        self.offsetWindow = None
        self.runWizardAtStartup = True

        self.processor.processStarted.connect(self.createProgressBar)
        self.processor.processFinished.connect(self.destroyProgressBar)
        self.processor.processUpdated.connect(self.updateProgressBar)


        self.setStyleSheet("QMainWindow::separator { background: rgb(200, 200, 200); width: 2px; height: 2px;}");

        self.configResource = ConfigResource("./test.xml")
        self.loadConfig()
        self.readSettings()
        self.tabify()

        self.show()
        if self.runWizardAtStartup:
            self.startWizard()

    def createActions(self):

        self.actionOpen = QAction('Open',self)
        self.actionSave_Session = QAction('Save project', self)
        self.actionOpen_Session = QAction('Open Session', self)
        self.actionExport_Wig = QAction('Export Wig', self)
        self.actionOffset = QAction('Set Offset', self)
        self.actionBatch = QAction('Batch...', self)
        self.actionWizard = QAction('Wizard', self)

        self.actionToggleTranscripts = QAction('Toggle Include', self)
        self.actionToggleGene = QAction('Toggle Include', self)

        self.menuFile = self.menubar.addMenu('File')
        self.menuFile.addAction(self.actionOpen)
        self.menuFile.addAction(self.actionOpen_Session)
        self.menuFile.addAction(self.actionSave_Session)
        self.menuFile.addAction(self.actionExport_Wig)
        self.menuFile.addAction(self.actionOffset)
        self.menuFile.addAction(self.actionBatch)
        self.menuFile.addAction(self.actionWizard)

        self.menuWindows = self.menubar.addMenu("Windows")


        self.toolbar.addAction(self.actionOpen)
        self.toolbar.addAction(self.actionExport_Wig)
        self.toolbar.addAction(self.actionSave_Session)
        self.toolbar.addAction(self.actionOffset)
        self.toolbar.addAction(self.actionWizard)

    def createMenus(self):
        self.menubar = QMenuBar(self)
        self.setMenuBar(self.menubar)
        self.statusbar = QStatusBar(self)
        self.setStatusBar(self.statusbar)

    def createToolbar(self):
        self.toolbar = QToolBar("Toolbar",self)
        self.addToolBar(self.toolbar)

    def createDockWidgets(self):
        self.setTabPosition(Qt.AllDockWidgetAreas, QTabWidget.North)
        self.mainplot = graphdockwidget('Plot', self)
        self.mainplot.setFeatures(QDockWidget.NoDockWidgetFeatures)

        self.resources = resourceDock(self)
        self.resources.setObjectName("Resources")
        self.datastats = statsdockwidget(self)

        self.addDockWidget(QtCore.Qt.LeftDockWidgetArea, self.resources)
        self.addDockWidget(QtCore.Qt.RightDockWidgetArea, self.mainplot)
        self.addDockWidget(QtCore.Qt.RightDockWidgetArea, self.datastats)

        self.dockTabList = [self.mainplot, self.datastats, self.resources]

    def createSignals(self):
        self.actionOffset.triggered.connect(self.createOffsetDock)
        self.actionOpen.triggered.connect(self.openFile)
        self.actionSave_Session.triggered.connect(self.saveConfigFile)
        self.actionOpen_Session.triggered.connect(self.openConfigFile)
        self.actionExport_Wig.triggered.connect(self.writeWigFile)
        self.actionBatch.triggered.connect(self.runBatch)
        self.actionWizard.triggered.connect(self.startWizard)


        self.resources.useTranscriptResource.connect(self.setActiveTranscriptRes)
        self.resources.useSequenceResource.connect(self.setActiveSequenceRes)
        self.resources.useNoiseResource.connect(self.setNoiseRes)

        self.datastats.refreshStats.connect(self.updateStats)
        self.datastats.filterData.connect(self.applyFilter)

    def updateMenu(self):
        self.menuWindows.clear()
        for a in  self.createPopupMenu().actions():
            self.menuWindows.addAction(a)

    def openFile(self):
        filename, _ = QFileDialog.getOpenFileName(self, "Open File", '',
                    "*.*")
        elements = filename.split('.')
        ext = elements[len(elements) - 1].lower()
        if ext == 'bam':
            #self.reader.openDatafile(filename)
            res = SequenceResource(filename)
            self.resources.append(res)

            res.processStarted.connect(self.createProgressBar)
            res.processFinished.connect(self.destroyProgressBar)
            res.processUpdated.connect(self.updateProgressBar)
            res.load()

        elif ext == 'bed' or ext == 'gtf':
            res = TranscriptResource(filename,ext)
            self.resources.append(res)

            res.processStarted.connect(self.createProgressBar)
            res.processFinished.connect(self.destroyProgressBar)
            res.processUpdated.connect(self.updateProgressBar)
            res.load()

        elif ext == 'xml':
            self.configResource = ConfigResource(filename)
            self.loadConfig()

    def loadFile(self, filename):
        elements = filename.split('.')
        ext = elements[len(elements) - 1].lower()
        if ext == 'bam':
            #self.reader.openDatafile(filename)
            res = SequenceResource(filename)
            self.resources.append(res)

            res.processStarted.connect(self.createProgressBar)
            res.processFinished.connect(self.destroyProgressBar)
            res.processUpdated.connect(self.updateProgressBar)
            res.load()

        elif ext == 'bed' or ext == 'gtf':
            res = TranscriptResource(filename,ext)
            self.resources.append(res)

            res.processStarted.connect(self.createProgressBar)
            res.processFinished.connect(self.destroyProgressBar)
            res.processUpdated.connect(self.updateProgressBar)
            res.load()

        elif ext == 'xml':
            self.configResource = ConfigResource(filename)
            self.loadConfig()

    def openConfigFile(self):

        filename, _ = QFileDialog.getOpenFileName(self, "Open Config File", '',
                    "Config files (*.xml)")
        self.configResource = ConfigResource(filename)
        self.loadConfig()

    def loadConfig(self):
        self.configResource.load()
        for name, res in self.configResource.resources.items():
            self.resources.append(res)

            res.processStarted.connect(self.createProgressBar)
            res.processFinished.connect(self.destroyProgressBar)
            res.processUpdated.connect(self.updateProgressBar)
            res.load()

        for (transcriptID, resourceName) in self.configResource.disabled:
            self.resources.getResource(resourceName).transcripts[transcriptID]['include'] = False

        if self.configResource.activeTranscript:
            self.resources.setSelectedResource(self.configResource.activeTranscript)

        if self.configResource.activeSequence:
            self.resources.setSelectedResource(self.configResource.activeSequence)

        if self.configResource.activeNoise:
            self.resources.setNoiceResource(self.configResource.activeNoise)

        self.currentRatio = self.configResource.currentRatio

        self.processor.lengthOffsets = self.configResource.lengthOffsets



    def saveConfigFile(self):
        filename, _ = QFileDialog.getSaveFileName(self, "Save Config File", '',
                    "Config files (*.xml)")
        if not filename:
            return
        self.configResource.filename = filename
        self.configResource.resources = self.resources.resources
        self.configResource.lengthOffsets = self.processor.lengthOffsets
        self.configResource.activeTranscript = self.activeTranscriptRes
        self.configResource.activeSequence = self.activeSequenceRes
        self.configResource.activeNoise = self.noiseResource
        if self.offsetWindow:
            self.currentRatio = self.offsetWindow.getRatio()
            self.configResource.currentRatio = self.currentRatio
        self.configResource.write()

    def writeWigFile(self):
        self.runBatch()

    def runBatch(self):
        #print(self.processor.lengthOffsets)
        self.dlg = batchDialog(self.resources.getResources(), self.processor.lengthOffsets)

        if self.dlg.exec_():

            useOffset, split, operation, lengths,relative, transcriptFilename, dataFileNames, outputDirectory = self.dlg.getConfig()
            #print(useOffset, split, operation, lengths, relative, transcriptFilename, dataFileNames, outputDirectory)
            if not transcriptFilename or not dataFileNames or not outputDirectory:
                return
            if self.resources.hasResource(transcriptFilename):
                transcriptRes = self.resources.getResource(transcriptFilename)
            else:
                ext = transcriptFilename.split('.')[-1]
                transcriptRes = TranscriptResource(transcriptFilename,ext)

            self.processor.setTranscripts(transcriptRes)

            for file in dataFileNames:


                if self.resources.hasResource(file):
                    dataResource = self.resources.getResource(file)
                else:
                    dataResource = SequenceResource(file)



                dataResource.setOffsets(self.processor.lengthOffsets)
                dataResource.setUseOffsets(useOffset)

                if not dataResource:
                    continue
                self.processor.setData(dataResource)

                filename = (file.split('/')[-1]).replace(".bam", "")
                
                if relative:
                    self.processor.writeRelativeWigFile(outputDirectory+'/'+filename, lengths, split, useOffset, 'ignore')
                else:
                    self.processor.writeWigFile(outputDirectory+'/'+filename, lengths, split)

    def startWizard(self):
        self.wiz = wizardDialog(self, self.resources.getResourcesDict(), self.runWizardAtStartup)
        self.wiz.show()

        self.wiz.finished.connect(self.endWizard)

    def endWizard(self, result):

        self.runWizardAtStartup = self.wiz.runAtStartup()
        if result:
            resTran = str(self.wiz.transResPage.selectedLabel.text())
            self.resources.setSelectedResource(self.resources.resources[resTran])

            resources = self.wiz.getResources()
            self.resources.clearSelection();
            if resources['sequence'] in self.resources.resources:
                self.resources.setSelectedResource(self.resources.resources[resources['sequence']])
            if resources['transcript'] in self.resources.resources:
                self.resources.setSelectedResource(self.resources.resources[resources['transcript']])
            if resources['noise'] in self.resources.resources:
                self.resources.setNoiceResource(self.resources.resources[resources['noise']])
            if self.wiz.calcOffset:
                self.createOffsetDock()

    def setOffset(self, ID, offset):
        if not(self.processor):
            return

        self.processor.setOffsets(ID, offset)

        try:
            self.datastats.updateOffsets( self.processor.lengthOffsets)
        except:
            return

    def dataLoaded(self,fileName):
        self.statusbar.showMessage("Data file loaded: " + fileName)

    def transcriptLoaded(self, fileName):
        self.statusbar.showMessage("Transcript file loaded: " + fileName)

    def configLoaded(self,fileName):
        self.setWindowTitle("Shoelaces - " + fileName)

    def transcriptSelected(self, transcriptList):

        self.selectedtranscripts = transcriptList

        self.mainplot.updateTranscriptStats(self.selectedtranscripts, self.processor)
        self.mainplot.show()
        self.mainplot.raise_()
        self.activateWindow()

    def geneSelected(self, geneList):

        if not geneList:
            self.selectedGene = None
            self.updateListView(False, True)
            self.mainplot.updateGeneStats(None, self.processor)
            return

        self.selectedGene = geneList[0]
        self.mainplot.updateGeneStats(self.selectedGene, self.processor)
        self.updateListView(False, True)

        self.mainplot.show()
        self.mainplot.raise_()
        self.activateWindow()

    def setActiveTranscriptRes(self, res):
        self.activeTranscriptRes = res

        if self.activeSequenceRes and self.activeTranscriptRes:
            self.activeTranscriptRes.addReadsToGene(self.activeSequenceRes)


        self.createListView()
        self.processor.setTranscripts(self.activeTranscriptRes)

    def setActiveSequenceRes(self, res):
        self.activeSequenceRes = res

        if self.activeTranscriptRes:
            self.activeTranscriptRes.addReadsToGene(self.activeSequenceRes)
            self.createListView()

        if not res:
            return



        if self.noiseResource:
            self.activeSequenceRes.setNoise(self.noiseResource)


        self.processor.setData(self.activeSequenceRes)

    def setNoiseRes(self, res):
        self.noiseResource = res

        if self.activeSequenceRes:
            self.activeSequenceRes.setNoise(self.noiseResource)

    def fileQuit(self):
        self.close()

    def closeEvent(self, ce):
        self.fileQuit()

    def createOffsetDock(self):
        if(not(self.processor)):
            return
        #print("creating offset dock")
        #self.reader.writeToFile(False)
        #return
        if self.offsetWindow:
            self.dockTabList.remove(self.offsetWindow)
            self.offsetWindow.close()
            self.removeDockWidget(self.offsetWindow)

        self.offsetWindow = offsetDock(self, self.processor,self.currentRatio)
        self.offsetWindow.setAllowedAreas(QtCore.Qt.LeftDockWidgetArea | QtCore.Qt.RightDockWidgetArea | QtCore.Qt.TopDockWidgetArea | QtCore.Qt.BottomDockWidgetArea)
        self.addDockWidget(QtCore.Qt.RightDockWidgetArea, self.offsetWindow)
        self.offsetWindow.emitUpdate.connect(self.setOffset)
        self.offsetWindow.resize(300,100)

        self.updateMenu()
        self.dockTabList.append( self.offsetWindow)

        self.tabify()
        self.offsetWindow.show()
        self.offsetWindow.raise_()
        self.offsetWindow.activateWindow()
        #self.setTabOrder(self.offsetWindow,self.dockTabList[0])

    def createListView(self):
        if not self.activeTranscriptRes:
            self.geneView.clear()
            self.transcriptView.clear()
            return
        if self.activeTranscriptRes.genes and not self.geneView:
            self.geneView = transcriptView('Genes',['Gene Name', '# transcripts', '# reads', 'Chromosome'], self)
            self.geneView.processStarted.connect(self.createProgressBar)
            self.geneView.processFinished.connect(self.destroyProgressBar)
            self.geneView.processUpdated.connect(self.updateProgressBar)
            self.geneView.selectionChanged.connect(self.geneSelected)
            self.addDockWidget(QtCore.Qt.LeftDockWidgetArea, self.geneView)

        if self.activeTranscriptRes and not self.transcriptView :
            self.transcriptView = transcriptView("Transcripts",['Transcript', 'Gene', 'Chromosome'],self)
            self.transcriptView.processStarted.connect(self.createProgressBar)
            self.transcriptView.processFinished.connect(self.destroyProgressBar)
            self.transcriptView.processUpdated.connect(self.updateProgressBar)
            self.transcriptView.view.setSelectionMode(QAbstractItemView.ExtendedSelection)
            self.transcriptView.selectionChanged.connect(self.transcriptSelected)
            self.addDockWidget(QtCore.Qt.LeftDockWidgetArea, self.transcriptView)

        self.geneView.usageToggled.connect(self.transcriptView.update)
        self.transcriptView.usageToggled.connect(self.geneView.update)
        self.updateListView()
        self.updateMenu()

    def updateListView(self, updateGeneView= True, updateTranscriptView = True ):
        if not self.activeTranscriptRes:
            return

        if updateGeneView and self.geneView:
            self.geneView.clear()

        if self.activeTranscriptRes.genes and updateGeneView:
            self.geneView.updateView(self.activeTranscriptRes.genes, ['gene_id','NoT', 'NoR','chromosome'])

        if updateTranscriptView and self.transcriptView:
            self.transcriptView.clear()
        if self.activeTranscriptRes.transcripts and updateTranscriptView:
            currentTranscripts = {}

            if self.selectedGene:
                currentTranscripts.update(self.selectedGene['transcripts'])
                currentTranscripts.update(self.selectedGene['non_coding'])
            else:
                currentTranscripts.update(self.activeTranscriptRes.transcripts)
                currentTranscripts.update(self.activeTranscriptRes.noncodingtranscripts)
            self.transcriptView.updateView(currentTranscripts, ['transcript_id', 'gene_id', 'chromosome'])

    def updateStats(self):

        if not self.activeSequenceRes or not self.activeTranscriptRes:
            return
        self.activeSequenceRes.setOffsets(self.processor.lengthOffsets)
        self.activeSequenceRes.setUseOffsets(self.datastats.useOffsets())
        self.activeSequenceRes.createWig()

        self.datastats.updateAvailableLengths(self.activeSequenceRes.splitWig['forward'].keys(), self.processor.lengthOffsets)
        self.datastats.updatestats(self.processor, 'global')

    def applyFilter(self):
        self.activeSequenceRes.filter()

    def createProgressBar(self, maxValue, name):


        self.progress = QProgressBar(self)
        self.progress.setRange(0,maxValue)
        self.statusbar.addWidget(self.progress)

        self.statusLabel = QLabel(self.statusbar)
        self.statusLabel.setText(name)
        self.statusbar.addWidget(self.statusLabel)
        self.progress.show()

    def destroyProgressBar(self, value):

        self.statusbar.removeWidget(self.progress)
        self.progress.destroy()

        self.statusbar.removeWidget(self.statusLabel)
        self.statusLabel.destroy()

    def tabify(self):
        first = self.dockTabList[0]
        for a in self.dockTabList[1:]:
            self.tabifyDockWidget(first,a)
            first = a

    def updateProgressBar(self,value):
        self.progress.setValue(value)

    def closeEvent(self,event):

        settings = QSettings("Bio", "Shoelaces")
        settings.setValue("geometry", self.saveGeometry())
        settings.setValue("windowState", self.saveState())
        settings.setValue("startupWizard",self.runWizardAtStartup);

        QMainWindow.closeEvent(self, event)

    def readSettings(self):

        settings = QSettings("Bio", "Shoelaces")
        if settings.value("geometry"):
            self.restoreGeometry(settings.value("geometry"))
        if settings.value("windowState"):
            self.restoreState(settings.value("windowState"))

        if settings.value("startupWizard") == None:
            self.runWizardAtStartup = True
        else:
            self.runWizardAtStartup = bool(settings.value("startupWizard"))



class commandLineApplication:

    def __init__(self, args):



        parser = argparse.ArgumentParser(description='Shoelaces: ribosome profiling analysis pipeline')

        parser.add_argument('-e', '--end', help="use CDS end (default: CDS start)", action="store_true")
        parser.add_argument('-o', '--offset', help="use offsets (Config xml file)", action="store_true")
        parser.add_argument('-a', '--automatic', help="detect offsets", action="store_true")
        parser.add_argument('-r', '--relative', help="output in transcript coordinates (default: genome coordinates)", action="store_true")

        parser.add_argument('-p', '--printplot', help="print out offset plot", action="store_true")
        parser.add_argument('-s', '--split', help="Split wig files based on read lengths", action="store_true")
        parser.add_argument('-lengths', metavar='N', type=int, nargs='*', help='output selected lengths into separate files')
        parser.add_argument('-files', metavar='F', type=str, nargs='*', help='[Config file, BAM file, GTF file, Output file]')

        args = parser.parse_args()

        if not args.files:
            sys.exit("no input file given")

        processor = sequenceProcessor()
        output = ''
        data = None
        transcript = None

        ratio = 1

        for filename in args.files:

            elements = filename.split('.')
            ext = elements[len(elements) - 1].lower()
            name = '.'.join(elements[:len(elements) - 1])
            print (filename, ext)
            if ext == 'xml':
                configRes = ConfigResource(filename)
                configRes.load()
                processor.lengthOffsets = configRes.lengthOffsets
                transcript = configRes.activeTranscript
                data = configRes.activeSequence
                ratio = configRes.currentRatio

            elif ext == 'gtf' or ext == 'bed' and not transcript:
                transcript = TranscriptResource(filename, ext)
            elif ext == 'bam' and not data:
                data = SequenceResource(filename)
            elif ext == 'wig':
                output = name

        if not(transcript):
            print("No active transcript resource")
            return
        if not(data):
            print("No active sequence resource")
            return
        if not output and not args.printplot:
            print("No output-file given")
            return
        transcript.load()
        processor.setTranscripts(transcript)
        processor.setData(data)


        if args.printplot:
            #print(ratio)
            transcript.addReadsToGene(data)
            result = processor.plotDifferenceToCDS(not(args.end), False, ratio)

            freqPlots =  processor.fourierTransform(result)
            common = processor.findUsefulLengths(freqPlots, result)

            filename = transcript.GetName()
            for length in common:
                freqPlot = freqPlots[length]

                f = open(filename + "-" + str(length) + ".frq", "w")
                for freq in sorted(freqPlot.keys()):
                    value = freqPlot[freq]
                    f.write(str(freq) + "\t" + str(value) +"\n")
                f.close()
                positions = sorted(result[length].keys())
                subwig = processor.getSubWig(result[length],positions, -15, 15)
                processor.populateHist(subwig,(-15,15))

                print('Sequence length:', length)
                for pos in sorted(subwig.keys()):
                    print(pos, subwig[pos])

            plots = {}

            for key, wig in result.items():
                hist = processor.getSubWig(wig, sorted(wig.keys()), -15, -5)
                plots[key] = hist

            offsets = processor.detectOffset(plots)
            print("Automatically detected offsets: ", offsets)
            return

        if args.automatic:
            transcript.addReadsToGene(data)
            result = processor.plotDifferenceToCDS(not(args.end), False, ratio)
            freqPlots =  processor.fourierTransform(result)
            common = processor.findUsefulLengths(freqPlots,result)
            plots = {}
            for key, wig in result.items():
                hist = processor.getSubWig(wig, sorted(wig.keys()), -15, -5)
                plots[key] = hist

            offsets = processor.detectOffset(plots)
            for key in common:
                processor.setOffsets(key, offsets[key])

            print("Automatically detected offsets", processor.lengthOffsets)




        offset = args.offset or args.automatic
        if offset:
            data.setUseOffsets(True)
        data.setOffsets(processor.lengthOffsets)
        data.createWig()

        #print("Lenghts: ", args.lengths)
        if not args.lengths:
            args.lengths = []
        split = args.lengths != [] or args.split


        if args.relative:
            processor.writeRelativeWigFile(output, args.lengths, split)
        else:
            processor.writeWigFile(output, args.lengths, split)



__author__ = 'birkeland'
from PyQt5.QtWidgets import *
from PyQt5 import QtCore
from PyQt5.QtCore import QSize

from shoelaces.helperwidgets import *
import shoelaces.settings as settings

class offsetWidget(QWidget):
    def __init__(self, parent = None):
        super(offsetWidget,self).__init__(parent)
        self.mainWidget = QScrollArea(self)
        self.mainWidget.setWidgetResizable(True)
        self.centralWidget = QMainWindow(self.mainWidget)
        self.centralWidget.setWindowFlags(QtCore.Qt.Widget)

        self.mainWidget.setWidget(self.centralWidget)

        self.mainLayout = QVBoxLayout(self)
        self.mainLayout.addWidget(self.mainWidget)
        self.widgets = []

    def addPlot(self, p, name):
        plotDock = QDockWidget(self.centralWidget)
        plotDock.setWidget(p)
        plotDock.setWindowTitle(name)
        self.centralWidget.addDockWidget(QtCore.Qt.LeftDockWidgetArea, plotDock)
        self.centralWidget.updateGeometry()
        p.parentDock = self


    def removeDock(self, dock):
        self.centralWidget.removeDockWidget(dock)


    #def updateSize(self):
    #    self.mainWidget.setWidget(self.centralWidget)

class offsetDock(QDockWidget):

    emitUpdate = pyqtSignal(int, int)

    def __init__(self, parent = None, processor = None, currentRatio = 0.1):

        super(offsetDock, self).__init__(parent)
        self.startplots = {}
        self.endplots = {}
        self.freqplots = {}
        self.modulusplots = {}
        self.updateRequried = False
        self.ratios = [1.0, 0.1, 0.05, 0.01]

        self.processor = processor

        self.mainWidget = QWidget(self)
        self.mainWidget.setMinimumSize(QSize(1000, 300))

        self.mainLayout = QGridLayout(self.mainWidget)

        self.percentage = QComboBox(self.mainWidget)
        self.percentage.addItem("All")
        self.percentage.addItem("Top 10 %")
        self.percentage.addItem("Top 5 %")
        self.percentage.addItem("Top 1 %")
        ind  = self.ratios.index(currentRatio)
        self.percentage.setCurrentIndex(ind)
        self.percentage.currentIndexChanged.connect(self.updatePlots)


        self.mainLayout.addWidget(self.percentage,1,0,1,2)


        self.tabsStart = QTabWidget(self.mainWidget)
        self.tabsEnd = QTabWidget(self.mainWidget)
        self.tabsFreq = QTabWidget(self.mainWidget)
        self.tabsModulus = QTabWidget(self.mainWidget)
        self.mainLayout.addWidget(self.tabsStart,3,0)
        self.mainLayout.addWidget(QLabel("Start codon"), 2, 0)
        self.mainLayout.addWidget(self.tabsEnd,3,1)
        self.mainLayout.addWidget(QLabel("Stop codon"), 2, 1)
        self.mainLayout.addWidget(self.tabsFreq,3,3)
        self.mainLayout.addWidget(QLabel("Reading frame"), 2, 2)
        self.mainLayout.addWidget(self.tabsModulus,3,2)
        self.mainLayout.addWidget(QLabel("Periodicity"), 2, 3)

        self.commonStartPlots = offsetWidget(self.tabsStart)
        self.otherStartPlots = offsetWidget(self.tabsStart)

        self.commonEndPlots = offsetWidget(self.tabsEnd)
        self.otherEndPlots = offsetWidget(self.tabsEnd)

        self.commonFreqGraphs = offsetWidget(self.tabsFreq)
        self.otherFreqGraphs = offsetWidget(self.tabsFreq)

        self.commonModulusGraphs = offsetWidget(self.tabsModulus)
        self.otherModulusGraphs = offsetWidget(self.tabsModulus)


        self.tabsStart.addTab(self.commonStartPlots, 'Common')
        self.tabsStart.addTab(self.otherStartPlots,'Others')

        self.tabsEnd.addTab(self.commonEndPlots, 'Common')
        self.tabsEnd.addTab(self.otherEndPlots,'Others')

        self.tabsFreq.addTab(self.commonFreqGraphs, 'Common')
        self.tabsFreq.addTab(self.otherFreqGraphs,'Others')

        self.tabsModulus.addTab(self.commonModulusGraphs, 'Common')
        self.tabsModulus.addTab(self.otherModulusGraphs,'Others')


        self.setWidget(self.mainWidget)


        #self.setRatio(currentRatio)

        self.createPlots()


        self.setWindowTitle("CDS Offset Plots")
        self.setObjectName("offset plots")

    def updatePlots(self, state):

        self.createPlots()

    def setRatio(self, ratio):


        ind  = self.ratios.index(ratio)
        self.percentage.setCurrentIndex(ind)
        self.updateRequried = True


    def getRatio(self):
        ratios = [1.0, 0.1, 0.05, 0.01]

        index = self.percentage.currentIndex()

        return ratios[index]

    def createPlots(self):


        if not self.processor:
            return
        self.percentage.hidePopup()

        ratios = [1.0, 0.1, 0.05, 0.01]

        index = self.percentage.currentIndex()

        result = self.processor.plotDifferenceToCDS(True, False, ratios[index])
        freqPlots =  self.processor.fourierTransform(result)
        modplots = self.processor.createModulusPlots(result)
        common = self.processor.findUsefulLengths(freqPlots, result)
        #print(common)
        #################
        #detect offsets
        subPlots =  {}
        for key, wig in result.items():
            subPlots[key] = self.processor.getSubWig(wig, sorted(wig.keys()), -15, -5)
        offsets = self.processor.detectOffset(subPlots)
        for key, offset in offsets.items():
            settings.getMainApp().setOffset(key, offset)
         #################
        #print(offsets)
        #print(common)

        self.createPlotsDocks(result, (-20,40),self.startplots, (self.commonStartPlots, self.otherStartPlots), common, glOffsetPlotWidget)

        result = self.processor.plotDifferenceToCDS(False, False, ratios[index])



        self.createPlotsDocks(result, (-40,20),self.endplots, (self.commonEndPlots, self.otherEndPlots), common, glOffsetPlotWidget, "stop")

        self.createPlotsDocks(freqPlots, (0,10),self.freqplots, (self.commonFreqGraphs, self.otherFreqGraphs), common, glGraphPlotWidget)
        self.createPlotsDocks(modplots, (0,3),self.modulusplots, (self.commonModulusGraphs, self.otherModulusGraphs), common, glModularPlotWidget)
        self.updateRequried = False

    def createPlotsDocks(self, result, interval, plotList, plotDocks, common, plotType, codon = "start"):


        for key in sorted(result.keys()):

            wig = result[key]
            hist = self.processor.getSubWig(wig, sorted(wig.keys()), interval[0], interval[1])

            plotParent = plotDocks[1]
            notParent = plotDocks[0]


            if key in common:
                plotParent =plotDocks[0]
                notParent = plotDocks[1]



            if key in plotList:

                 plotList[key].updatePlot(hist)
                 if plotList[key].parentDock != plotParent:
                     dock = plotList[key].parent()
                     plotParent.addPlot(plotList[key],str(key))
                     notParent.removeDock(dock)

                 continue

            plot = plotType(plotParent, key, hist, codon)
            plotParent.addPlot(plot, str(key))
            plotList[key] = plot
            plot.offsetUpdated.connect(self.emitUpdate)
            plot.offsetUpdated.connect(self.updateOffset)

            if key in self.processor.lengthOffsets:
                plot.setOffset(self.processor.lengthOffsets[key])

        plotDocks[0].show()
        plotDocks[1].show()



    def updateOffset(self, key, offset):
        try:
            if self.sender().codon == 'start':
                self.endplots[self.sender().ID].setOffset(offset)
                self.modulusplots[self.sender().ID].setOffset(offset)
            elif self.sender().codon == 'stop':
                self.startplots[self.sender().ID].setOffset(offset)
                self.modulusplots[self.sender().ID].setOffset(offset)
            else:
                self.startplots[self.sender().ID].setOffset(offset)
                self.endplots[self.sender().ID].setOffset(offset)
        except :
            print ("error")

    def sizeHint(self):
        return QSize(500, 500)


from PyQt5.QtWidgets import *
from PyQt5.QtCore import *
import numpy as np

from shoelaces.readProcessing import *
from shoelaces.changepoint_analysis import *


class sequenceProcessor(QObject):

    configLoaded = pyqtSignal(str)

    processStarted = pyqtSignal(int, str)
    processUpdated = pyqtSignal(int)
    processFinished = pyqtSignal(int)

    def __init__(self, transcripts = None, sequenceData = None, noise = None):
        QObject.__init__(self)
        self.transcripts = transcripts
        self.sequenceData = sequenceData
        self.lengthOffsets = {}
        self.noise = noise
        self.noiseIntervals = {}
    # result index
    #0: Chromoson    #1: RNA start coord    #2: RNA end coord    #3: name    #4: direction    #5: CDS start coord    #6: CDS end coord    #7: exon sizes    #8: exon starts

    def __bool__(self):
        if not self.transcripts:
            return False
        if not self.sequenceData:
            return False
        return True

    def setData(self, sequenceData):
        self.sequenceData = sequenceData

    def setTranscripts(self, transcripts):
        self.transcripts = transcripts

    def getReadBySpecs(self, pos, cigarstring, chromosome):
        reads = self.sequenceData.fetch(chromosome, pos, pos+1)
        for read in reads:
            if read.pos == pos and read.cigarstring == cigarstring:
                return read

    def checkOverlap(self, coord, read, checkToStart):
        threshold = 30
        pos = read.pos

        if read.is_reverse:
            if checkToStart:
                pos += read.alen - 1
                addRead = coord < pos < coord + threshold
            else:

                addRead = coord > pos > coord - threshold
        else:
            if checkToStart:
                addRead = coord > pos > coord - threshold
            else:
                pos += read.alen

                addRead = coord < pos < coord + threshold
        return addRead


    def plotDifferenceToCDS(self, cds_start = True, useOffsets = False, percentage = 1.0, ):
        if not self.sequenceData or not self.transcripts:
            return {}

        readBins = {}

        readsPertranscript = {}
        totalReads = 0

        usedCDScoords= []

        NoT = len(self.transcripts.transcripts)*percentage
        #print("Used transcripts:",NoT)
        progress = 0;

        self.processStarted.emit(NoT+1, "Processing Sequence Data")
        transcriptList =   self.transcripts.transcripts.items()

        if self.transcripts.genes:
            #print("Sorting transcripts")
            transcriptList = reversed(sorted(self.transcripts.transcripts.items(), key = lambda t: float(self.transcripts.genes[t[1]['gene_id']]['NoR'])/float(t[1]['CDS_length']) ))
        for (transcriptName, transcript) in transcriptList:

            #print(transcript[0])
            if progress >= NoT:
                break
            if not transcript['include']:
               # print('Not using', transcript[1]['transcript_id'])
                continue

            if not 'stop_codon' in transcript or not 'start_codon' in transcript:
                continue

            self.processUpdated.emit(progress)
            progress += 1

            forward = transcript['forward']

            interval = (-50, 200)

            if (cds_start and forward) or (not cds_start and not forward):
                coord = transcript['CDS_min']
            else:
                coord = transcript['CDS_max']
                interval = (-200, 50)

            #print(transcript[0], coord, interval)

            if self.sequenceData.isNoise(coord):
                continue

            if coord in usedCDScoords:
                continue

            chromosome = transcript['chromosome']
            fetchStartCoord = moveOnTranscript(coord, transcript, interval[0])
            fetchEndCoord = moveOnTranscript(coord, transcript, interval[1])
            reads = self.sequenceData.fetch(chromosome, fetchStartCoord, fetchEndCoord)

            usedCDScoords.append(coord)
            validReads = 0
            #minPos = 1000000000000
            #maxPos = 0
            for read in reads:
                if isMultiMapper(read):
                    continue

                validReads += 1

                #minPos = min(minPos, getRealReadPos(read,0))
                #maxPos = max(maxPos, getRealReadPos(read,0))

                length = read.infer_query_length()
                try:
                    readBins[length].append((read,  transcriptName))
                except:
                    readBins[length] = [(read,  transcriptName)]
            readsPertranscript[transcriptName] = validReads
            totalReads += validReads
            # print("transcript: ", transcriptName)
            # print("span: " , fetchEndCoord - fetchEndCoord)
            # print("read span: ", maxPos - minPos)

        nrOfPlots = len(readBins)
        plots = {}
        normalizedWeights = {}

        for key in readsPertranscript.keys():

            normalizedWeights[key] =  1.0 #/ readsPertranscript[key]

        self.processFinished.emit(0)
        QApplication.processEvents()

        self.processStarted.emit(nrOfPlots, "Generating Difference Plots")

        for key, i in zip(readBins, range(0,nrOfPlots)):
            self.processUpdated.emit(i)

            diffs = {}

            offset = 0
            # if useOffsets and key in self.lengthOffsets:
            #     offset = self.lengthOffsets[key]
            # elif useOffsets:
            #     continue

            for (read, transcriptId) in readBins[key]:
                cds_min = True
                if (read.is_reverse and cds_start) or (not cds_start and not read.is_reverse):
                    cds_min = False

                diff = getDiffToCDS(read, self.transcripts[transcriptId], offset, cds_min)

                diff += offset

                try:
                    diffs[diff] += normalizedWeights[transcriptId]
                except:
                    diffs[diff] = normalizedWeights[transcriptId]

            plots[key] = diffs
            sortedplot = sorted(diffs.items(), key=lambda x: x[0])
            # if key==28:
            #     for key in sortedplot:
            #         print(key)
        self.processFinished.emit(0)
        QApplication.processEvents()
        return plots

    def fourierTransform(self, plots):

        result = {}

        for length, plot in plots.items():

            self.populateHist(plot,range(1,150))

            sortedplot = sorted(plot.items(), key=lambda x: x[0])

            data = np.array([(val) for (key,val) in sortedplot if key >= 0 and key <= 150],np.float )
            try:
                ft = np.fft.fft(data, axis=0)
            except:
                continue
            amplitudes = abs(ft)
            frequencies = np.fft.fftfreq(len(data), 1)
            hist = {1/i:value for i, value in zip(frequencies, amplitudes) if i != 0}
            result[length] = hist
        return result

    def createModulusPlots(self, plots, offset = 1):
        results = {}
        for length, plot in plots.items():
            modplot = {}

            for pos, value in plot.items():
                m = offset + (pos % 3)
                try:
                    modplot[m] += plot[pos]
                except:
                    modplot[m] = plot[pos]
            results[length] = modplot
        return results

    def detectOffset(self, plots):
        result = {}
        self.processStarted.emit(len(plots.keys()), "Detecting Read Offsets")
        process = 0
        for length, plot in plots.items():
            self.processUpdated.emit(process)
            process+=1
            self.populateHist(plot, range(-30, 10))
            sortedplot = sorted(plot.items(), key=lambda x: x[0])


            #print(sortedplot)
            data = np.array([(val) for (key,val) in sortedplot],np.float )
            try:
                stats, pvals, nums = detect_mean_shift(data)
            except:
                continue

            changes = np.zeros(len(stats)-1)
            for s in range(len(stats)-1):
                changes[s] = stats[s+1] -  stats[s]


            ind = np.argmax(abs(changes))+1


            #print(ind)
            result[length] = sortedplot[ind][0]

        self.processFinished.emit(0)
        return result

    def findUsefulLengths(self, freqPlots, plots):

        usefulLengths = []
        for key,plot in freqPlots.items():

            largest = 0.0
            freq = 0.0
            func = self.getSubWig(plot,sorted(plot.keys()),1,15)

            plot = plots[key]
            sum= 0;
            for i, value in plot.items():
                sum += value;
            if sum <100:
                print("Not enough reads")
                continue;

            for currentFreq, value in func.items():

                if value > largest:

                    largest = value
                    freq = abs(currentFreq)

            if abs(freq - 3.0) < 0.1:# and largest/stdDev > 3:
                usefulLengths.append(key)

        return usefulLengths

    def populateHist(self, hist, range):
        for i in range:
            try:
                hist[i]
            except KeyError:
                hist[i] = 0

    def detectReadOffsets(self, data):

        for plot in data:

            if len(plot[0].keys()) < 15:
                continue
            length = plot[1]
            graph = sorted(plot[0])
            index = 0
            maxDiff = graph[0]
            for i in range(1, len(graph)):
                diff = (graph[i] - graph[i-1])
                if diff > maxDiff:
                    index = i
                    maxDiff = diff
            self.lengthOffsets[length] = -index

        #print self.lengthOffsets

    def setOffsets(self, readLength, offsets):

        self.lengthOffsets[readLength] = offsets

    #cigar
    #
    #M	BAM_CMATCH	    0
    #I	BAM_CINS	    1
    #D	BAM_CDEL	    2
    #N	BAM_CREF_SKIP	3
    #S	BAM_CSOFT_CLIP	4
    #H	BAM_CHARD_CLIP	5
    #P	BAM_CPAD	    6
    #=	BAM_CEQUAL	    7
    #X	BAM_CDIFF	    8

    def plotSumOntranscript(self, transcriptName, useOffsets=False):

        if not self.sequenceData:
            return {}
        if not(transcriptName in self.transcripts):
            #print "Not existing transcript:", transcriptName
            return {}

        transcript = self.transcripts[transcriptName]
        fetchStart = transcript['RNA_min']
        fetchEnd = transcript['RNA_max']
        chromosome = transcript['chromosome']

        reads = self.sequenceData.fetch(chromosome, fetchStart, fetchEnd)

        return self.plotsum(reads)

    def plotReadsOntranscript(self, transcriptName, useOffsets = False):

        if not self.sequenceData:
            return {}
        if not(transcriptName in self.transcripts.transcripts):
            #print "Not existing transcript:", transcriptName
            return {}

        transcript = self.transcripts[transcriptName]
        fetchStart = transcript['RNA_min']
        fetchEnd = transcript['RNA_max']

        chromosome = transcript['chromosome']
        try:
            reads = self.sequenceData.fetch(chromosome, fetchStart, fetchEnd)
        except ValueError:
            reads = []

        return self.plotReads(reads, useOffsets)

    def plotSum(self, reads):

        readPositions = {}
        count =0
        for read in reads:
            #length, fills, gaps = parseRead(read)



            for (start, finish) in read.get_blocks():
                for i in range(start+1, finish+1):
                    if(i == 24068169):
                        #print(read.qname)
                        #print(getRealReadPos(read,0))
                        count += 1;
                    try:
                        readPositions[i] += 1
                    except KeyError:
                        readPositions[i] = 1

        return readPositions

    def plotReads(self, reads, useOffsets = False):

        if not self.sequenceData:
            return {}

        readPositions = {}

        for read in reads:
            length, fill, gaps = parseRead(read)

            offset = 0
            if useOffsets and length in self.lengthOffsets:
                offset = self.lengthOffsets[length]
            elif useOffsets:
                continue

            #nucPos = getModulatedReadPos(read, transcript,offset)
            nucPos = getRealReadPos(read, offset)

            #if nucPos == -1:
                #print read.cigar, read.pos

            if nucPos in readPositions:
                readPositions[nucPos] += 1
            else:
                readPositions[nucPos] = 1

        return readPositions

    def gettranscriptStats(self, transcriptName, useOffsets=False, multimappers = 'ignore'):


        if not self.sequenceData:
            print ("No data")
            return {}

        if not (transcriptName in self.transcripts.transcripts):
            print ("Not existing transcript:", transcriptName)
            return {}

        transcript = self.transcripts[transcriptName]
        #print(transcript)

        fetchStart = transcript['RNA_min']
        fetchEnd = transcript['RNA_max']
        chromosome = transcript['chromosome']


        smallestCDS = transcript['CDS_min']
        largestCDS = transcript['CDS_max']

                #stop_codon not included in cds
        if('stop_codon' in transcript):
            if(transcript['forward']):
                largestCDS = transcript['stop_codon'][1]
            else:
                smallestCDS = transcript['stop_codon'][0]


        reads = self.sequenceData.fetch(chromosome, fetchStart, fetchEnd)
        #print ('Nr of reads:', len(reads))

        onUTRmin = 0
        onCDS = 0
        onUTRmax = 0
        isMM = 0
        onIntron = 0
        ignoreMM  = multimappers == 'ignore'



        for read in reads:

            mm = isMultiMapper(read)
            if mm and ignoreMM:
                isMM += 1
                continue

            length, fill, gaps = parseRead(read)

            offset = 0
            if useOffsets and length in self.lengthOffsets:
                offset = self.lengthOffsets[length]
            elif useOffsets:

                continue

            exonIndex = getOnExon(read, transcript, offset)

            if exonIndex == -2:

                continue
            elif exonIndex == -1:
                onIntron +=1
                continue

            pos = getRealReadPos(read, offset)
            if pos < smallestCDS:
                onUTRmin += 1
            elif pos > largestCDS:
                onUTRmax += 1
            elif smallestCDS <=  pos <= largestCDS:
                onCDS += 1



        if not transcript['forward']:
            temp = onUTRmax
            onUTRmax = onUTRmin
            onUTRmin = temp

        result = {}
        result['CDS'] = onCDS
        result['Leader'] = onUTRmin
        result['Trailer'] = onUTRmax
        result['Intron'] = onIntron
        result['Multimappers'] = isMM

        result['Sum'] =result['CDS']+result['Leader']+result['Trailer']+result['Intron']+result['Multimappers']
        return result

    def getgenestats(self, gene, useOffsets= False, multimappers = 'ignore'):
        if not self.sequenceData:
            return {}

        fetchStart = gene['RNA_min']
        fetchEnd = gene['RNA_max']
        chromosome = gene['chromosome']

        #smallestCDS = transcript['CDS_min']
        #largestCDS = transcript['CDS_max']
        if gene['reads']:
            reads = gene['reads']
        else:
            reads = self.sequenceData.fetch(chromosome, fetchStart, fetchEnd)
            gene['reads'] = reads

        onUTRmin = 0
        onCDS = 0
        onUTRmax = 0
        onIntron = 0
        onNoneCoding = 0
        isMM = 0
        ignoreMM = multimappers =='ignore'
        #print("Use offsets: ", useOffsets)
        for read in reads:

            mm =  isMultiMapper(read)

            if mm and ignoreMM:
                isMM += 1
                continue

            length, fill, gaps = parseRead(read)
            offset = 0
            if useOffsets and length in self.lengthOffsets:
                offset = self.lengthOffsets[length]
            elif useOffsets:
                continue

            inCDS = False
            inUTRmin = False
            inUTRmax = False
            inIntron = False
            for transcriptName, transcript in gene['transcripts'].items():

                smallestCDS = transcript['CDS_min']
                largestCDS = transcript['CDS_max']
                if('stop_codon' in transcript):
                    if(transcript['forward']):
                        largestCDS = transcript['stop_codon'][1]
                    else:
                        smallestCDS = transcript['stop_codon'][0]
                exonIndex = getOnExon(read, transcript, offset)

                if exonIndex == -2:
                    continue
                elif exonIndex == -1:
                    inIntron = True
                    continue

                pos = getRealReadPos(read, offset)
                if largestCDS >= pos >= smallestCDS :
                    inCDS = True
                    break
                if pos > largestCDS:
                    inUTRmax = True
                if pos < smallestCDS:
                    inUTRmin = True



            if inCDS:
                onCDS += 1
            elif inUTRmax:
                onUTRmax += 1
            elif inUTRmin:
                onUTRmin += 1
            elif inIntron:
                onIntron += 1
            else:
                for transcriptName, transcript in self.transcripts.noncodingtranscripts.items():
                    exonIndex = getOnExon(read, transcript, offset)
                    if exonIndex:
                        onNoneCoding += 1
                        break



        if not gene['forward']:
            temp = onUTRmax
            onUTRmax = onUTRmin
            onUTRmin = temp


        result = {}
        result['CDS'] = onCDS
        result['Leader'] = onUTRmin
        result['Trailer'] = onUTRmax
        result['Intron'] = onIntron
        result['Non-coding'] = onNoneCoding
        result['Multimappers'] = isMM
        result['Sum'] =result['CDS']+result['Leader']+result['Trailer']+result['Intron']+result['Non-coding']+result['Multimappers']
        return result

    def getglobalstats(self, useOffsets= False, multimappers = 'ignore'):

        return {}
        if not self.sequenceData:
            return {}

        onUTRmin = 0
        onCDS = 0
        onUTRmax = 0
        isMM = 0
        onIntron = 0
        onNoneCoding = 0
        ignoreMM = multimappers =='ignore'

        progress = 0;

        progressBar = QProgressDialog("Processing Sequence Data", "Abort", 0, len(self.transcripts.chromosomes))
        #progressBar.setWindowModality(Qt.WindowModal)
        progressBar.show()


        transcriptList = sorted(self.transcripts.transcripts.items(), key = lambda t: t[1]['RNA_min'])

        for chr in self.transcripts.chromosomes:
            #print("Fetching reads")
            reads = self.sequenceData.fetch(chr)

            i = 0
            for read in reads:

                i+=1
                mm =  isMultiMapper(read)

                if mm and ignoreMM:
                    isMM += 1
                    continue

                length, fill, gaps = parseRead(read)
                offset = 0
                if useOffsets and length in self.lengthOffsets:
                    offset = self.lengthOffsets[length]
                elif useOffsets:
                    continue

                inCDS = False
                inUTRmin = False
                inUTRmax = False
                inIntron = False

                pos = getRealReadPos(read, offset)

                for transcriptName, transcript in self.transcripts.transcripts.items():
                    smallestCDS = transcript['CDS_min']
                    largestCDS = transcript['CDS_max']
                    exonIndex = getOnExon(read, transcript, offset)

                    if exonIndex == -2:
                        continue
                    elif exonIndex == -1:
                        inIntron = True
                        continue

                    if largestCDS >= pos >= smallestCDS :
                        inCDS = True
                        break
                    if (pos > largestCDS and transcript['forward']) or (pos<smallestCDS and not transcript['forward']):
                        inUTRmax = True
                    else:
                        inUTRmin = True


                if inCDS:
                    onCDS += 1
                elif inUTRmax:
                    onUTRmax += 1
                elif inUTRmin:
                    onUTRmin += 1
                elif inIntron:
                    onIntron += 1


            progress +=1
            progressBar.setValue(progress)
            QApplication.processEvents()
        progressBar.close()
        result = {}
        result['CDS'] = onCDS
        result['Leader'] = onUTRmin
        result['Trailer'] = onUTRmax
        result['Intron'] = onIntron
        result['Multimappers'] = isMM


        return result

    def countStats(self, wig, forward ):

        onUTRmin = 0
        onCDS = 0
        onUTRmax = 0
        isMM = 0
        onIntron = 0
        onNoneCoding = 0
        sumReads = 0

        for chr, histogram in wig.items():

            positions = histogram
            if not chr in self.transcripts.chromosomes:
                continue

            transcriptList = self.transcripts.chromosomes[chr]
            for pos, count in positions.items():
                sumReads += count
                inCDS = False
                inUTRmin = False
                inUTRmax = False
                inIntron = False

                for transcript in transcriptList:
                    if transcript['forward'] != forward or not transcript['include']:
                        continue

                    smallestCDS = transcript['CDS_min']
                    largestCDS = transcript['CDS_max']
                    if('stop_codon' in transcript):
                        if(transcript['forward']):
                            largestCDS = transcript['stop_codon'][1]
                        else:
                            smallestCDS = transcript['stop_codon'][0]

                    exonIndex = getPosOnExon(pos, transcript)

                    if exonIndex == -2:
                        continue
                    elif exonIndex == -1:
                        inIntron = True
                        continue

                    if largestCDS >= pos >= smallestCDS :
                        inCDS = True
                        break
                    if (pos > largestCDS and transcript['forward']) or (pos<smallestCDS and not transcript['forward']):
                        inUTRmax = True
                    else:
                        inUTRmin = True


                if inCDS:
                    onCDS += count
                elif inUTRmax:
                    onUTRmax += count
                elif inUTRmin:
                    onUTRmin += count
                elif inIntron:
                    onIntron += count
                else:
                    for transcriptName, transcript in self.transcripts.noncodingtranscripts.items():
                        exonIndex = getPosOnExon(pos, transcript)
                        if exonIndex:
                            onNoneCoding += count
                            break
        return onUTRmin, onCDS, onUTRmax, onIntron, onNoneCoding, sumReads

    def getglobalstatsFromWig(self, useOffsets, lengths = []):

        if not self.sequenceData:
            return {}
        if not self.transcripts:
            return {}


        self.sequenceData.setUseOffsets(useOffsets)
        self.sequenceData.createWig()
        if not useOffsets and not lengths:
            lengths = self.sequenceData.splitWig['forward'].keys()

        elif(not lengths):
            lengths = self.sequenceData.splitWig['forward'].keys()
            lengthDist = {}

            for l in lengths :
                lengthDist[l] = 0


            result = {}
            result['Selected lengths'] = 'None'
            result['CDS'] = 0
            result['Leader'] = 0
            result['Trailer'] = 0
            result['Intron'] =  0
            result['Non-coding'] = 0
            result['Multimappers'] = 0
            return result, lengthDist



        onUTRminf, onCDSf, onUTRmaxf, onIntronf, onNoneCodingf = 0,0,0,0,0
        onUTRminb, onCDSb, onUTRmaxb, onIntronb, onNoneCodingb = 0,0,0,0,0
        lengthDist = {}

        self.processStarted.emit(len(lengths), "Processing Wig")
        counter = 0
        for length in lengths:
            self.processUpdated.emit(counter)
            counter += 1
            lengthDist[length] = 0
            try:
                a,b,c,e,f, sum =  self.countStats(self.sequenceData.splitWig['forward'][length], True)
                onUTRminf+=a
                onCDSf+=b
                onUTRmaxf+=c

                onIntronf+=e
                onNoneCodingf+=f
                lengthDist[length] += sum

            except KeyError:
                #print('Wrong forward length')
                pass

            try:

                g,h,i,k,l, sum =  self.countStats(self.sequenceData.splitWig['reverse'][length], False)
                onUTRminb += g
                onCDSb+=h
                onUTRmaxb+=i

                onIntronb+=k
                onNoneCodingb+=l
                lengthDist[length] += sum

            except KeyError:
                #print('Wrong reverse length')
                pass
        self.processFinished.emit(counter )

        result = {}



        result['CDS'] = onCDSf + onCDSb
        result['Leader'] = onUTRminf + onUTRminb
        result['Trailer'] = onUTRmaxf + onUTRmaxb
        result['Intron'] = onIntronf + onIntronb
        result['Non-coding'] = onNoneCodingf + onNoneCodingb
        result['Multimappers'] = self.sequenceData.multimappers
        result['Total amount of reads'] = result['CDS'] + result['Leader'] + result['Trailer'] + result['Intron'] + result['Non-coding'] + result['Multimappers']

        return result, lengthDist

    def plotAlltranscripts(self, function, transcripts = []):
        if not self.sequenceData:
            return {}
        if not transcripts:
            transcripts = self.transcripts.keys()
        result = {}
        for transcript in transcripts:
            #if (self.transcripts[transcript]['forward']):
            #    continue
            currentResult = function(transcript)
            for key in currentResult:
                if key in result:
                    result[key] += currentResult[key]
                else:
                    result[key] = currentResult[key]
        return result


    def writeWigToFile(self, filename, wig):
        print(filename)
        #print(wig)
        wigFile = open(filename, 'w')
        for chromosome, wig in wig.items():
            print(chromosome)
            wigFile.write("variableStep chrom=" + chromosome+"\n")
            for pos in sorted(wig):
                line = str(pos) + '\t' + str(wig[pos]) + '\n'
                wigFile.write(line)
        wigFile.close()


    def writeWigFile(self, filename, selectedlengths = [], split = False):


        if not self.sequenceData:
            return


        self.sequenceData.createWig()

        if split:
            if selectedlengths == []:
                forwardKeys = self.sequenceData.splitWig['forward'].keys()
                reverseKeys = self.sequenceData.splitWig['reverse'].keys()

                selectedlengths = list(set(forwardKeys).union(set(reverseKeys)))
            for length in selectedlengths:
                outputfile = filename + '-' + str(length)+'-forward.wig'
                try:
                    self.writeWigToFile(outputfile, self.sequenceData.splitWig['forward'][length])
                except KeyError:
                    continue
                outputfile = filename + '-' + str(length)+'-reverse.wig'
                try:
                    self.writeWigToFile(outputfile, self.sequenceData.splitWig['reverse'][length])
                except KeyError:
                    continue
        else:
            outputfile = filename +'-forward.wig'
            if self.sequenceData.forwardWig:
                self.writeWigToFile(outputfile, self.sequenceData.forwardWig)

            outputfile = filename +'-reverse.wig'
            if self.sequenceData.reverseWig:
                self.writeWigToFile(outputfile, self.sequenceData.reverseWig)



    def getSubWig(self, wig, positions, start, end):
        result = {}

        for pos in positions:
            #print(pos)
            if pos > end:
                break

            elif pos >= start:
                result[pos] = wig[pos]



        return result


    def writeRelateWigToFile(self, filename, wig):


        wigFile = open(filename, 'w')
        for chromosome, relateiveWig in wig.items():

            for transcriptName, wig in relateiveWig.items():
                #print(chromosome, transcriptName)
                wigFile.write("variableStep chrom=" + str(chromosome)+ " transcript_id="+transcriptName+"\n")
                for pos in sorted(wig):
                    line = str(pos) + '\t' + str(wig[pos]) + '\n'
                    wigFile.write(line)
        wigFile.close()

    def createRelativeWig(self, wig, forward):
        relativeWig = {}

        for chromosome, currentWig in wig.items():
            currentRelWig = {}

            positions = sorted(currentWig.keys())
            #print("Creating subwig chr:", chromosome)
            for transcriptName in self.transcripts.transcripts:
                transcript = self.transcripts[transcriptName]
                if transcript['chromosome'] != chromosome:
                    continue

                subWig = self.getSubWig(currentWig, positions, transcript['RNA_min'], transcript['RNA_max'])
                #print("Subwig", subWig)
                relSubWig = {}
                for pos, value in subWig.items():
                    transcriptpos = dnatotranscript(pos, transcript)
                    relSubWig[transcriptpos] = value
                currentRelWig[transcriptName] = relSubWig
            relativeWig[chromosome] = currentRelWig

        return relativeWig


    def writeRelativeWigFile(self, filename, selectedlengths, split = False, useOffset = False, multimappers = 'ignore'):

        if not self.sequenceData:
            return


        self.sequenceData.createWig()

        if split:

            if selectedlengths == []:
                forwardKeys = self.sequenceData.splitWig['forward'].keys()
                reverseKeys = self.sequenceData.splitWig['reverse'].keys()

                selectedlengths = list(set(forwardKeys).union(set(reverseKeys)))
            for length in selectedlengths:
                outputfile = filename + '-' + str(length)+'-forward.wig'
                try:
                    relWig = self.createRelativeWig(self.sequenceData.splitWig['forward'][length], True)
                    self.writeRelateWigToFile(outputfile, relWig)
                except KeyError:
                    continue
                outputfile = filename + '-' + str(length)+'-reverse.wig'
                try:
                    relWig = self.createRelativeWig(self.sequenceData.splitWig['reverse'][length], False)
                    self.writeRelateWigToFile(outputfile, relWig)
                except KeyError:
                    continue
        else:
            outputfile = filename +'-forward.wig'
            if self.sequenceData.forwardWig:
                relWig = self.createRelativeWig( self.sequenceData.forwardWig, True)
                self.writeRelateWigToFile(outputfile, relWig)

            outputfile = filename +'-reverse.wig'
            if self.sequenceData.reverseWig:
                relWig = self.createRelativeWig( self.sequenceData.reverseWig, False)
                self.writeRelateWigToFile(outputfile, relWig)









import numpy as np

def detect_mean_shift( ts, B=1000):
    """ Detect mean shift in a time series. B is number of bootstrapped
        samples to draw.
    """
    x = np.arange(0, len(ts))
    stat_ts_func = compute_balance_mean_ts
    null_ts_func = shuffle_timeseries
    stats_ts, pvals, nums = get_ts_stats_significance(x, ts, stat_ts_func, null_ts_func, B=B, permute_fast=True)
    return stats_ts, pvals, nums

def compute_balance_mean( ts, t):
    """ Compute the balance. The right end - the left end."""
    """ For changed words we expect an increase in the mean, and so only 1 """
    return np.mean(ts[t + 1:]) - np.mean(ts[:t + 1])

def compute_balance_mean_ts( ts):
    """ Compute the balance at each time 't' of the time series."""
    balance = [compute_balance_mean(ts, t) for t in np.arange(0, len(ts) - 1)]
    return balance

def shuffle_timeseries( ts):
    """ Shuffle the time series. This also can serve as the NULL distribution. """
    return np.random.permutation(ts)

def get_ts_stats_significance( x, ts, stat_ts_func, null_ts_func, B=1000, permute_fast=False, label_ts=''):
    """ Returns the statistics, pvalues and the actual number of bootstrap
        samples. """
    stats_ts, pvals, nums = ts_stats_significance(
        ts, stat_ts_func, null_ts_func, B=B, permute_fast=permute_fast)
    return stats_ts, pvals, nums


def ts_stats_significance(ts, ts_stat_func, null_ts_func, B=1000, permute_fast=False):
    """ Compute  the statistical significance of a test statistic at each point
        of the time series.
    """
    stats_ts = ts_stat_func(ts)
    if permute_fast:
        # Permute it in 1 shot
        null_ts = map(np.random.permutation, np.array([ts, ] * B))
    else:
        null_ts = np.vstack([null_ts_func(ts) for i in np.arange(0, B)])
    stats_null_ts = np.vstack([ts_stat_func(nts) for nts in null_ts])
    pvals = []
    nums = []
    for i in np.arange(0, len(stats_ts)):
        num_samples = np.sum((stats_null_ts[:, i] >= stats_ts[i]))
        nums.append(num_samples)
        pval = num_samples / float(B)
        pvals.append(pval)

    return stats_ts, pvals, nums